import copy
from collections import defaultdict
from pathlib import Path
from typing import DefaultDict, Dict, Generator, List, Set, Union

import spacy
import srsly
from spacy.language import Language
from spacy.scorer import Scorer
from tqdm import tqdm, tqdm_notebook
from wasabi import Printer

from .constants import NONE
from .recognizer import EntityRecognizer
from .types import (
    AnnotationCount,
    Example,
    HardestExample,
    LabelDisparity,
    PredictionError,
    PredictionErrorExamplePair,
    Span,
)


def get_ents_by_label(
    data: List[Example], case_sensitive: bool = False
) -> DefaultDict[str, List[str]]:
    """Get a dictionary of unique text spans by label for your data

    # TODO: Ok so this needs to return more than just a set for each label.

    We want to return a dictionary that maps labels to AnnotationCount objects where each 
    AnnotationCount contains the text of the annotation text, the total number of times it's mentioned (e.g. what entity_coverage does)
    but also the examples it is in. 

    So maybe I can get this info from entity_coverage? IDK but this is dumb rn and not very flexible.

    Maybe I should keep this function returning a set of strings for each label for compatability but I need the other way too
    so I know what to focus on in editing and a 
    
    Args:
        data (List[Example]): List of examples
        case_sensitive (bool, optional): Consider case of text for each annotation
        sort_by (SortBy): Sort by text or by count
    
    Returns:
        DefaultDict[str, List[str]]: DefaultDict mapping label to sorted list of the unique
            spans annotated for that label.
    """
    annotations: DefaultDict[str, Set[str]] = defaultdict(set)
    sorted_annotations: DefaultDict[str, List[str]] = defaultdict(list)

    for e in data:
        for s in e.spans:
            span_text = s.text if case_sensitive else s.text.lower()
            annotations[s.label].add(span_text)

    for label, anns in annotations.items():
        sorted_annotations[label] = sorted(anns)

    return sorted_annotations


def get_label_disparities(
    data: List[Example], label1: str, label2: str, case_sensitive: bool = False
) -> Set[str]:
    """Identify annotated spans that have different labels in different examples
    
    Args:
        data (List[Example]): Input List of examples
        label1 (str): First label to compare
        label2 (str): Second label to compare
        case_sensitive (bool, optional): Consider case of text for each annotation
    
    Returns:
        Set[str]: Set of all unique text spans that overlap between label1 and label2
    """
    annotations = get_ents_by_label(data, case_sensitive=case_sensitive)
    return set(annotations[label1]).intersection(set(annotations[label2]))


def top_label_disparities(
    data: List[Example], case_sensitive: bool = False, dedupe: bool = False
) -> List[LabelDisparity]:
    """Identify annotated spans that have different labels
    in different examples for all label pairs in data.
    
    Args:
        data (List[Example]): Input List of examples
        case_sensitive (bool, optional): Consider case of text for each annotation
        dedupe (bool, optional): Whether to deduplicate for table view vs confusion matrix.
            False by default for easy confusion matrix display.
    
    Returns:
        List[LabelDisparity]: List of LabelDisparity objects for each label pair combination
            sorted by the number of disparities between them.
    """
    annotations = get_ents_by_label(data, case_sensitive=case_sensitive)
    label_disparities = {}
    for label1 in annotations.keys():
        for label2 in annotations.keys():
            if label1 != label2:
                intersection = set(annotations[label1]).intersection(set(annotations[label2]))
                n_disparities = len(intersection)
                if n_disparities > 0:
                    if dedupe:
                        input_hash = "||".join(sorted([label1, label2]))
                    else:
                        input_hash = "||".join([label1, label2])

                    label_disparities[input_hash] = LabelDisparity(
                        label1=label1, label2=label2, count=n_disparities
                    )

    return sorted(label_disparities.values(), key=lambda ld: ld.count, reverse=True)


def top_prediction_errors(
    recognizer: EntityRecognizer,
    data: List[Example],
    labels: List[str] = None,
    n: int = None,
    k: int = None,
    exclude_fp: bool = False,
    exclude_fn: bool = False,
    verbose: bool = False,
) -> List[PredictionError]:
    """Get a sorted list of examples your model is worst at predicting.
    
    Args:
        recognizer (EntityRecognizer): An instance of EntityRecognizer
        data (List[Example]): List of annotated Examples
        labels (List[str], optional): List of labels to get errors for. 
            Defaults to the labels property of `recognizer`.
        n (int, optional): If set, only use the top n examples from data.
        k (int, optional): If set, return the top k prediction errors, otherwise the whole list.
        exclude_fp (bool, optional): Flag to exclude False Positive errors.
        exclude_fn (bool, optional): Flag to exclude False Negative errors.
        verbose (bool, optional): Show verbose output.
    
    Returns:
        List[PredictionError]: List of Prediction Errors your model is making, sorted by the
            spans your model has the most trouble with.
    """
    labels_ = labels or recognizer.labels
    if n is not None:
        data = data[:n]

    n_examples = len(data)
    texts = (e.text for e in data)
    anns = (e.spans for e in data)

    errors = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))  # type: ignore
    error_examples: DefaultDict[str, List[PredictionErrorExamplePair]] = defaultdict(list)
    n_errors = 0

    for orig_example, pred_example, ann in zip(data, recognizer.predict(texts), anns):
        if k is not None and n_errors > k:
            break

        pred_error_example_pair = PredictionErrorExamplePair(
            original=orig_example, predicted=pred_example
        )

        cand = set([(s.start, s.end, s.label) for s in pred_example.spans])
        gold = set([(s.start, s.end, s.label) for s in ann])

        fp_diff = cand - gold
        fn_diff = gold - cand

        seen = set()

        if fp_diff and not exclude_fp:
            for fp in fp_diff:
                gold_ent = None
                for ge in gold:
                    if fp[0] == ge[0] and fp[1] == ge[1]:
                        gold_ent = ge
                        break
                if gold_ent:
                    start, end, label = gold_ent
                    text = pred_example.text[start:end]
                    false_label = fp[2]
                    errors[label][text][false_label] += 1
                    error_examples[f"{text}||{label}||{false_label}"].append(
                        pred_error_example_pair
                    )
                else:
                    start, end, false_label = fp
                    text = pred_example.text[start:end]
                    errors[NONE][text][false_label] += 1
                    error_examples[f"{text}||{NONE}||{false_label}"].append(pred_error_example_pair)
                n_errors += 1
                seen.add((start, end))

        if fn_diff and not exclude_fn:
            for fn in fn_diff:
                start, end, label = fn
                if (start, end) not in seen:
                    text = pred_example.text[start:end]
                    errors[label][text][NONE] += 1
                    error_examples[f"{text}||{label}||{NONE}"].append(pred_error_example_pair)
                    n_errors += 1

    ranked_errors_map: Dict[str, PredictionError] = {}

    for label, errors_per_label in errors.items():
        for error_text, error_labels in errors_per_label.items():
            for error_label, count in error_labels.items():
                pe_hash = f"{error_text}||{label}||{error_label}"
                ranked_errors_map[pe_hash] = PredictionError(
                    text=error_text,
                    true_label=label,
                    pred_label=error_label,
                    count=count,
                    examples=error_examples[f"{error_text}||{label}||{error_label}"],
                )

    ranked_errors: List[PredictionError] = sorted(
        list(ranked_errors_map.values()), key=lambda error: error.count, reverse=True  # type: ignore
    )
    error_texts = set()
    for re in ranked_errors:
        if re.examples:
            for e in re.examples:
                error_texts.add(e.original.text)

    error_rate = round(len(error_texts) / len(data), 2)
    if verbose:
        error_summary = {
            "N Examples": len(data),
            "N Errors": len(ranked_errors),
            "N Error Examples": len(error_texts),
            "Error Rate": error_rate,
        }
        msg = Printer()
        msg.divider("Error Analysis")
        msg.table(error_summary)

    return ranked_errors


def get_hardest_examples(
    pred_errors: List[PredictionError],
    return_pred_errors: bool = True,
    remove_pred_error_examples: bool = True,
) -> List[HardestExample]:
    """Get hardest examples from list of PredictionError types
    
    Args:
        pred_errors (List[PredictionError]): list of PredictionError
        return_pred_errors (bool, optional): Whether to return prediction errors. Defaults to True.
        remove_pred_error_examples (bool, optional): Whether to remove examples from returned PredictionError. Defaults to True.
    
    Raises:
        ValueError: Each PredictionError must have a List of examples 
    
    Returns:
        List[HardestExample]: Sorted list of the hardest examples for a model to work on.
    """

    has_examples = any([pe.examples for pe in pred_errors])
    if not has_examples:
        raise ValueError(
            "Each PredictionError in Parameter pred_errors must have examples attached."
        )

    examples_text_map: Dict[str, Example] = {}
    example_pred_errors_map: DefaultDict[str, List[PredictionError]] = defaultdict(list)
    for pe in pred_errors:
        if pe.examples:
            for example in pe.examples:
                examples_text_map[example.original.text] = example.original
                example_pred_errors_map[example.original.text].append(
                    PredictionError(
                        text=pe.text,
                        true_label=pe.true_label,
                        pred_label=pe.pred_label,
                        count=pe.count,
                        examples=[example],
                    )
                )

    hardest_examples = []
    for example_text, example_pred_errors in example_pred_errors_map.items():
        example = examples_text_map[example_text]  # type: ignore

        prediction_errors: List[PredictionError] = []
        if remove_pred_error_examples and example_pred_errors:
            prediction_errors = [
                PredictionError(
                    text=pe.text,
                    true_label=pe.true_label,
                    pred_label=pe.pred_label,
                    count=pe.count,
                    examples=[],
                )
                for pe in example_pred_errors
            ]
        else:
            prediction_errors = example_pred_errors

        prediction_error_hashes: Set[str] = set()
        deduped_prediction_errors: List[PredictionError] = []

        for pe in prediction_errors:
            pe_hash = f"{pe.text}||{pe.true_label}||{pe.pred_label}"
            if pe_hash not in prediction_error_hashes:
                deduped_prediction_errors.append(pe)

        record = HardestExample(example=example, count=len(deduped_prediction_errors))
        if return_pred_errors:
            record.prediction_errors = deduped_prediction_errors
        hardest_examples.append(record)

    sorted_hardest_examples = sorted(hardest_examples, key=lambda he: he.count, reverse=True)
    return sorted_hardest_examples



from pathlib import Path

import srsly
import typer
from recon.corpus import Corpus
from recon.stats import get_ner_stats


def main(data_dir: Path):
    corpus = Corpus.from_disk(data_dir)
    corpus_stats = corpus.apply(get_ner_stats, serialize=True)
    for name, stats in corpus_stats.items():
        print(f"{name}")
        print("=" * 50)
        print(stats)


if __name__ == "__main__":
    typer.run(main)



# isort:skip_file
# type: ignore

from collections import Counter, defaultdict
import copy
import random
from typing import Dict, Iterable, List, Optional, Union

import catalogue
import prodigy
from prodigy.components.db import connect
from prodigy.components.loaders import get_stream
from prodigy.components.preprocess import add_tokens
from prodigy.recipes.ner import get_labels_from_ner
from prodigy.util import (
    INPUT_HASH_ATTR,
    TASK_HASH_ATTR,
    get_labels,
    log,
    set_hashes,
    split_string,
)
import spacy
import srsly
from wasabi import msg
from recon.constants import NONE
from recon.dataset import Dataset
from recon.operations import op_iter, operation, Operation
from recon.types import HardestExample, Example, Span
from recon.validation import remove_overlapping_entities


def make_span_hash(span: Union[Span, Dict[str, object]]) -> str:
    if isinstance(span, Span):
        h = f"{span.text}|||{span.start}|||{span.end}|||{span.label}"
    else:
        h = f'{span["text"]}|||{span["start"]}|||{span["end"]}|||{span["label"]}'
    return h


def get_stream_from_hardest_examples(nlp, hardest_examples: List[HardestExample]):
    for he in hardest_examples:
        task = he.example.dict()
        gold_span_hashes = {make_span_hash(span): span for span in he.example.spans}
        predicted_example = None
        assert he.prediction_errors is not None
        for pe in he.prediction_errors:
            assert pe.examples is not None
            for e in pe.examples:
                if e.predicted.text == he.example.text:
                    predicted_example = e.predicted
        if predicted_example:
            pthtml = []
            predicted_example_task = predicted_example.dict()

            for token in predicted_example.tokens:
                pthtml.append(f'<span class="recon-token">{token.text} </span>')

            pred_spans = predicted_example.spans
            pred_span_hashes = [make_span_hash(span) for span in predicted_example.spans]

            for gold_span_hash, gold_span in gold_span_hashes.items():
                if gold_span_hash not in pred_span_hashes:
                    gold_span_miss = copy.deepcopy(gold_span)
                    gold_span_miss.label = NONE
                    pred_spans.append(gold_span_miss)

            pred_spans = remove_overlapping_entities(
                sorted([span.dict() for span in pred_spans], key=lambda s: s["start"])
            )

            i = len(pred_spans) - 1
            while i >= 0:
                span = pred_spans[i]
                span_hash = make_span_hash(span)
                if span_hash in gold_span_hashes:
                    labelColorClass = "recon-pred-success-mark"
                elif span["label"] == NONE:
                    labelColorClass = "recon-pred-missing-mark"
                else:
                    labelColorClass = "recon-pred-error-mark"

                if not span["token_end"]:
                    print(span)

                pthtml = (
                    pthtml[: span["token_end"]]
                    + [
                        f'<span class="recon-pred-label">{span["label"]}<span class="c0178">x</span></span></span>'
                    ]
                    + pthtml[span["token_end"] :]
                )
                pthtml = (
                    pthtml[: span["token_start"]]
                    + [f'<span class="recon-pred {labelColorClass}">']
                    + pthtml[span["token_start"] :]
                )
                i -= 1

            task[
                "html"
            ] = f"""
            <h2 class='recon-title'>Recon Prediction Errors</h2>
            <h5 class='recon-subtitle'>
                The following text shows the errors your model made on this example inline.
                Correct the annotations above based on how well your model performed on this example.
                If your labeling is correct you might need to add more training examples in this domain.    
            </h5>
            <div class='recon-container'>
                {''.join(pthtml)}
            </div>
            """
        task["prediction_errors"] = [pe.dict() for pe in he.prediction_errors]
        yield task


@prodigy.recipe(
    "recon.ner_correct",
    dataset=("Dataset to save annotations to", "positional", None, str),
    spacy_model=(
        "Base model or blank:lang (e.g. blank:en) for blank model",
        "positional",
        None,
        str,
    ),
    hardest_examples=(
        "Data to annotate (file path or '-' to read from standard input)",
        "positional",
        None,
        str,
    ),
    label=(
        "Comma-separated label(s) to annotate or text file with one label per line",
        "option",
        "l",
        get_labels,
    ),
    exclude=(
        "Comma-separated list of dataset IDs whose annotations to exclude",
        "option",
        "e",
        split_string,
    ),
)
def ner_correct(
    dataset: str,
    spacy_model: str,
    hardest_examples: List[HardestExample],
    label: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
):
    """
    Stream a List of `recon.types.HardestExample` instances to prodigy
    for review/correction. Uses the Prodigy blocks interface to display
    prediction error information along with ner view
    """
    log("RECIPE: Starting recipe recon.ner_correct", locals())
    if spacy_model.startswith("blank:"):
        nlp = spacy.blank(spacy_model.replace("blank:", ""))
    else:
        nlp = spacy.load(spacy_model)
    labels = label  # comma-separated list or path to text file
    if not labels:
        labels = get_labels_from_ner(nlp)
        if not labels:
            msg.fail("No --label argument set and no labels found in model", exits=1)
        msg.text(f"Using {len(labels)} labels from model: {', '.join(labels)}")

    log(f"RECIPE: Annotating with {len(labels)} labels", labels)
    labels = labels.split(",")

    stream = get_stream_from_hardest_examples(nlp, hardest_examples)
    # stream = add_tokens(nlp, stream)  # add "tokens" key to the tasks
    # stream = set_hashes(stream)

    table_template = """
    <style>
        .table-title {
            margin-top: -120px;
        }
        table {
            border-collapse: collapse;
        }
        table, td, th {
            border: 1px solid gray;
            vertical-align: top;
            margin-top: -50px;
        }
    </style>
    <h5 class='recon-subtitle table-title'>All prediction errors for this example.</h2>
    <table>
        <tr>
            <th>Text</th>
            <th>True Label</th>
            <th>Pred Label</th>
        </tr>
        {{#prediction_errors}}
        <tr>
            <td>{{text}}</td>
            <td>{{true_label}}</td>
            <td>{{pred_label}}</td>
        </tr>
        {{/prediction_errors}}
    </table>
    """

    return {
        "view_id": "blocks",
        "dataset": dataset,
        "stream": stream,
        "exclude": exclude,
        "config": {
            "lang": nlp.lang,
            "labels": labels,
            "exclude_by": "input",
            "blocks": [
                {"view_id": "ner_manual"},
                {"view_id": "html"},
                {"view_id": "html", "field_rows": 3, "html_template": table_template},
            ],
            "global_css": f"""
            .prodigy-container {{
                max-width: {'900px' if len(labels) > 8 else '600px'}
            }}
            .recon-title {{
                text-align: left;
                margin-top: -80px;
            }}
            .recon-subtitle {{
                text-align: left;
                margin-top: -80px;
                white-space: normal;
            }}
            .recon-container {{
                text-align: left;
                line-height: 2;
                margin-top: -80px;
                white-space: pre-line;
            }}
            .recon-pred {{
                color: inherit;
                margin: 0 0.15em;
                display: inline;
                padding: 0.25em 0.4em;
                font-weight: bold;
                line-height: 1;
                -webkit-box-decoration-break: clone;
            }}
            .recon-pred-success-mark {{
                background: #00cc66;
            }}
            .recon-pred-error-mark {{
                background: #fc7683;
            }}
            .recon-pred-missing-mark {{
                background: #84b4c4;
            }}
            .recon-pred-label {{
                color: #583fcf;
                font-size: 0.675em;
                font-weight: bold;
                font-family: "Roboto Condensed", "Arial Narrow", sans-serif;
                margin-left: 8px;
                text-transform: uppercase;
                vertical-align: middle;
            }}
            """,
        },
    }


@prodigy.recipe(
    "recon.ner_merge",
    # fmt: off
    dataset=("Dataset with saved annotations to from recon.ner_correct", "positional", None, str),
    recon_dataset=("Recon dataset name", "positional", None, str),
    source=("Source data to merge examples with (file path or List of examples)", "positional", None, str),
    output_dir=("Optional output directory to save dataset to", "positional", None, str),
    exclude=("Comma-separated list of dataset IDs whose annotations to exclude", "option", "e", split_string),
    # fmt: on
)
def ner_merge(
    dataset: str,
    recon_dataset: str,
    source: Union[str, Dataset],
    output_dir: Optional[str] = None,
    exclude: Optional[List[str]] = None,
):
    """
    Stream a List of `recon.types.HardestExample` instances to prodigy
    for review/correction. Uses the Prodigy blocks interface to display
    prediction error information along with ner view
    """
    log("RECIPE: Starting recipe recon.ner_merge", locals())
    if isinstance(source, str):
        dataset = Dataset(recon_dataset).from_disk(source)
    else:
        dataset = source

    DB = connect()
    if dataset not in DB:
        msg.fail(f"Can't find dataset '{dataset}'", exits=1)

    prodigy_raw_examples = DB.get_dataset(dataset)
    prodigy_examples = [Example(**eg) for eg in prodigy_raw_examples if eg["answer"] == "accept"]
    prodigy_texts_to_examples = {e.text: e for e in prodigy_examples}

    prev_len = len(dataset)
    dataset.apply_("recon.v1.prodigy.merge_examples", prodigy_texts_to_examples)
    assert len(dataset) == prev_len

    if output_dir:
        log(f"RECIPE: Fixing {len(prodigy_examples)} examples in data")
        dataset.to_disk(output_dir)


@operation("recon.v1.prodigy.merge_examples")
def merge_examples(example, prodigy_texts_to_examples):
    if example.text in prodigy_texts_to_examples:
        return prodigy_texts_to_examples[example.text]
    else:
        return example



import shutil
from pathlib import Path

import pytest
from recon.corpus import Corpus
from recon.stats import get_ner_stats

# def test_corpus_initialize(example_data):
#     corpus1 = Corpus(example_data["train"], example_data["dev"])
#     assert len(corpus1.train) == 106
#     assert len(corpus1.dev) == 110
#     assert len(corpus1.all) == 216

#     corpus2 = Corpus(example_data["train"], example_data["dev"], test=example_data["test"])
#     assert len(corpus2.train) == 106
#     assert len(corpus2.dev) == 110
#     assert len(corpus2.test) == 96
#     assert len(corpus2.all) == 312


# def test_corpus_disk(example_data):
#     corpus = Corpus(example_data["train"], example_data["dev"], test=example_data["test"])
#     save_dir = Path(__file__).parent.parent / "examples/data/skills/to_disk_test"

#     corpus.to_disk(save_dir, force=True)
#     corpus2 = Corpus.from_disk(save_dir)
#     assert corpus.train == example_data["train"]
#     assert corpus.dev == example_data["dev"]
#     assert corpus.test == example_data["test"]

#     assert corpus2.train == example_data["train"]
#     assert corpus2.dev == example_data["dev"]
#     assert corpus2.test == example_data["test"]
#     assert len(corpus.all) == len(corpus2.all)

#     shutil.rmtree(save_dir)


# def test_corpus_apply(example_data):
#     corpus = Corpus(example_data["train"], example_data["dev"], test=example_data["test"])
#     stats = corpus.apply(get_ner_stats)
#     assert sorted(list(stats.keys())) == ["all", "dev", "test", "train"]

#     assert stats["all"].n_examples == 312


# def test_corpus_apply_inplace(example_data):
#     corpus = Corpus(example_data["train"], example_data["dev"], test=example_data["test"])
#     stats = corpus.apply(get_ner_stats)
#     assert sorted(list(stats.keys())) == ["all", "dev", "test", "train"]

#     assert stats["all"].n_examples == 312



from recon.insights import (
    get_hardest_examples,
    top_label_disparities,
    top_prediction_errors,
)


def test_top_label_disparities(example_corpus, example_corpus_processed):
    top_disparities = top_label_disparities(example_corpus.all)
    assert len(top_disparities) == 8

    top_disparities = top_label_disparities(example_corpus.all, dedupe=True)
    assert len(top_disparities) == 4
    assert top_disparities[0].count == 1

    top_disparities_processed = top_label_disparities(example_corpus_processed.all)
    assert len(top_disparities_processed) == 2

    top_disparities_processed = top_label_disparities(example_corpus_processed.all, dedupe=True)
    assert len(top_disparities_processed) == 1
    assert top_disparities_processed[0].count == 2


def test_top_prediction_errors(recognizer, example_corpus):
    pred_errors = top_prediction_errors(recognizer, example_corpus.test)
    assert len(pred_errors) == 67


def test_get_hardest_examples(recognizer, example_corpus):
    pred_errors = top_prediction_errors(recognizer, example_corpus.test)
    hardest_examples = get_hardest_examples(pred_errors)

    assert len(hardest_examples) == 42
    assert hardest_examples[0].example.text.startswith("Some of the free Apache Tomcat resources")



from pathlib import Path
from typing import Dict, List

import pytest
from recon.corpus import Corpus
from recon.loaders import read_jsonl
from recon.preprocess import SpacyPreProcessor
from recon.recognizer import SpacyEntityRecognizer
from recon.types import Example, TransformationCallbacks
from spacy.lang.en import English


@pytest.fixture()
def nlp():
    return English()


@pytest.fixture()
def spacy_preprocessor(nlp):
    return SpacyPreProcessor(nlp)


@pytest.fixture()
def test_texts():
    return [
        "Machine learning is the most researched area of AI.",
        "My title at work is Software Engineer even theough I mostly work on AI.",
    ]


@pytest.fixture()
def example_data() -> Dict[str, List[Example]]:
    """Fixture to load example train/dev/test data that has inconsistencies.
    
    Returns:
        Dict[str, List[Example]]: Dataset containing the train/dev/test split
    """
    base_path = Path(__file__).parent.parent / "examples/data/skills"
    return {
        "train": read_jsonl(base_path / "train.jsonl"),
        "dev": read_jsonl(base_path / "dev.jsonl"),
        "test": read_jsonl(base_path / "test.jsonl"),
    }


@pytest.fixture()
def example_corpus() -> Corpus:
    """Fixture to load example train/dev/test data that has inconsistencies.
    
    Returns:
        Corpus: Example data
    """
    base_path = Path(__file__).parent.parent / "examples/data/skills"
    return Corpus.from_disk(base_path)


@pytest.fixture()
def example_corpus_processed() -> Corpus:
    """Fixture to load example train/dev/test data that has inconsistencies.
    
    Returns:
        Corpus: Example data
    """
    base_path = Path(__file__).parent.parent / "examples/data/skills"
    corpus = Corpus.from_disk(base_path)
    corpus.pipe_(
        [
            "recon.v1.fix_tokenization_and_spacing",
            "recon.v1.add_tokens",
            "recon.v1.upcase_labels",
            "recon.v1.filter_overlaps",
        ]
    )
    return corpus


@pytest.fixture()
def recognizer(nlp, example_corpus):
    patterns = []

    for e in example_corpus.all:
        for span in e.spans:
            patterns.append({"label": span.label, "pattern": span.text.lower()})

    nlp.add_pipe(nlp.create_pipe("entity_ruler", {"patterns": patterns}))

    recognizer = SpacyEntityRecognizer(nlp)
    return recognizer



from typing import Any, Dict, Iterable, List

import spacy
from spacy.language import Language
from spacy.tokens import Doc

from .types import Example


class PreProcessor(object):
    name = "recon.v1.preprocess"

    def __init__(self) -> None:
        super().__init__()
        self._cache: Dict[Any, Any] = {}

    def __call__(self, data: List[Example]) -> Iterable[Any]:
        raise NotImplementedError


class SpacyPreProcessor(PreProcessor):
    name = "recon.v1.spacy"

    def __init__(self, nlp: Language) -> None:
        super().__init__()
        self._nlp = nlp

    def __call__(self, data: List[Example]) -> Iterable[Any]:
        unseen_texts = (e.text for i, e in enumerate(data) if hash(e) not in self._cache)
        seen_texts = ((i, e.text) for i, e in enumerate(data) if hash(e) in self._cache)

        docs = list(self._nlp.pipe(unseen_texts))
        for doc in docs:
            self._cache[doc.text] = doc
        for idx, st in seen_texts:
            docs.insert(idx, self._cache[st])

        return docs



from datetime import datetime
from enum import Enum
from typing import Any, Callable, Dict, List, NamedTuple, Optional, Tuple, cast

from pydantic import BaseModel, Field, Schema, root_validator

from .hashing import example_hash, span_hash, token_hash, tokenized_example_hash


class Span(BaseModel):
    """Entity Span in Example"""

    text: str
    start: int
    end: int
    label: str
    token_start: Optional[int]
    token_end: Optional[int]

    def __hash__(self) -> int:
        return cast(int, span_hash(self))


class Token(BaseModel):
    """Token with offsets into Example Text"""

    text: str
    start: int
    end: int
    id: int

    def __hash__(self) -> int:
        return cast(int, token_hash(self))


class Example(BaseModel):
    """Example with NER Label spans"""

    text: str
    spans: List[Span]
    tokens: Optional[List[Token]]
    meta: Dict[str, Any] = {}
    formatted: bool = False

    @root_validator(pre=True)
    def span_text_must_exist(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if not values.get("formatted", False):
            # Ensure each span has a text property
            spans = values["spans"]
            for span in spans:
                if not isinstance(span, Span):
                    if "text" not in span:
                        span["text"] = values["text"][span["start"] : span["end"]]

            # Ensure the meta has a source property
            # if something that's not a dict is passed in
            meta = values.get("meta", {})
            if isinstance(meta, list) or isinstance(meta, str):
                meta = {"source": meta}

            values["spans"] = spans
            values["meta"] = meta
            values["formatted"] = True

        return values

    def __hash__(self) -> int:
        return cast(int, tokenized_example_hash(self))


class TransformationType(str, Enum):
    EXAMPLE_ADDED = "EXAMPLE_ADDED"
    EXAMPLE_REMOVED = "EXAMPLE_REMOVED"
    EXAMPLE_CHANGED = "EXAMPLE_CHANGED"


class Transformation(BaseModel):
    prev_example: Optional[int] = None
    example: Optional[int] = None
    type: TransformationType


# fmt: off
def add_shim(example: Example) -> None:
    return None

def remove_shim(example_hash: int) -> None:
    return None

def change_shim(example_hash: int, new_example: Example) -> None:
    return None

def track_shim(example_hash: Optional[int] = None, new_example: Optional[Example] = None) -> None:
    return None
# fmt: on


class TransformationCallbacks(NamedTuple):
    add_example: Callable[[Example], None] = add_shim
    remove_example: Callable[[int], None] = remove_shim
    change_example: Callable[[int, Example], None] = change_shim
    track_example: Callable[[Optional[int], Optional[Example]], None] = track_shim


class OperationStatus(str, Enum):
    NOT_STARTED = "NOT_STARTED"
    IN_PROGRESS = "IN_PROGRESS"
    COMPLETED = "COMPLETED"


class OperationState(BaseModel):
    name: str
    batch: bool = False
    args: List[Any] = []
    kwargs: Dict[str, Any] = {}
    status: OperationStatus = OperationStatus.NOT_STARTED
    ts: datetime = datetime.now()
    examples_added: int = 0
    examples_removed: int = 0
    examples_changed: int = 0
    transformations: List[Transformation] = []


class DatasetOperationsState(BaseModel):
    name: str
    commit: str
    size: int
    operations: List[OperationState]


class OperationResult(BaseModel):
    data: Any
    state: OperationState


class CorpusApplyResult(BaseModel):
    train: Any
    dev: Any
    test: Any
    all: Any

    def items(self) -> List[Tuple[str, Any]]:
        return [("train", self.train), ("dev", self.dev), ("test", self.test), ("all", self.all)]


class AnnotationCount(BaseModel):
    text: str
    count: int
    examples: List[Example]


class PredictionErrorExamplePair(BaseModel):
    """Dataclass representation of original Example in a PredictionError
    vs the predicted Example from an EntityRecognizer 
    
    Attributes:
        original (Example): Original Example
        spans (List[Span]): List of entity spans
        meta (Dict[str, Any], optional): Meta information about the example
    """

    original: Example
    predicted: Example


class PredictionError(BaseModel):
    """Representation of errors an EntityRecognizer makes on a labeled dataset.
    
    Attributes:
        text (str): Span text with error
        true_label (str): True label in annotated Example
        pred_label (str): The label predicted by the EntityRecognizer
        count (int): Number of times this PredictionError occurs
        examples (List[PredictionErrorExamplePair], optional):
            List of PredictionErrorExamplePairs that have this PredictionError
    """

    text: str
    true_label: str
    pred_label: str
    count: int
    examples: Optional[List[PredictionErrorExamplePair]] = []


class HardestExample(BaseModel):
    """Container for how hard an Example is for an EntityRecognizer to predict
    all entities correctly for.
    """

    example: Example
    count: int
    prediction_errors: Optional[List[PredictionError]]


class LabelDisparity(BaseModel):
    """Container for the number of disparities in a Dataset
    where some text is tagged as label1 in some places and label2 in others
    
    Attributes:
        label1 (str): Label1
        label2 (str): Label2
        count (int): Number of times this label disparity occurs
        examples (List[Example], optional): List of examples where this disparity occurs
    """

    label1: str
    label2: str
    count: int
    examples: Optional[List[Example]] = []


class NERStats(BaseModel):
    """Container for tracking basic NER statistics"""

    n_examples: int
    n_examples_no_entities: int
    n_annotations: int
    n_annotations_per_type: Dict[str, int]
    examples_with_type: Optional[Dict[str, Example]]


class EntityCoverage(BaseModel):
    """Container for tracking how well an Entity is covered.
    
    Attributes:
        text (str): The entity text
        label (str): The entity label
        count (int): Number of times this text/label combination occurs
        examples (List[Example], optional): List of examples where this entity occurs
    """

    text: str
    label: str
    count: int
    examples: Optional[List[Example]] = []

    def __hash__(self) -> int:
        return hash((self.text, self.label))


class EntityCoverageStats(BaseModel):
    """Container for output of how similar the Entity Coverage of 2 datasets is"""

    entity: float
    count: float


class Outliers(BaseModel):
    """Container for low and high indices of outlier detection"""

    low: List[int]
    high: List[int]


class Correction(BaseModel):
    """Container for an annotation correction, mapping an annotation from a label to a label"""

    annotation: str
    from_labels: List[str]
    to_label: str



from recon.operations import Operation, operation, registry


def test_operation_init():
    @operation("test_operation")
    def operation_test(example):
        return example

    assert "test_operation" in registry.operations
    assert isinstance(registry.operations.get("test_operation"), Operation)



from recon.operations import op_iter
from recon.preprocess import SpacyPreProcessor
from recon.tokenization import add_tokens
from recon.types import Example


def test_add_tokens(spacy_preprocessor):
    # fmt: off
    untokenized_examples = [Example(**example) for example in [
        {"text": "Have you used the new version of my model?", "spans": [{"start": 36, "end": 41, "label": "SKILL"}]},
        {"text": "I'd like to work as an actor or model if possible.", "spans": [{"text": "actor", "start": 23, "end": 28, "label": "JOB_ROLE"}, {"text": "model", "start": 32, "end": 37, "label": "JOB_ROLE"}]},
        {"text": "We are looking for a Software Development Engineer who has solid coding skills, a strong machine learning background, and is passionate about developing new AI products.", "spans": [{"start": 21, "end": 50, "label": "SKILL"}, {"start": 65, "end": 71, "label": "SKILL"}, {"start": 89, "end": 105, "label": "SKILL"}, {"start": 142, "end": 152, "label": "SKILL"}, {"start": 157, "end": 159, "label": "SKILL"}]},
        {"text": "Responsibilities As a SOFTWARE DEVELOPMENT ENGINEER II you will work / collaborate with other talented engineers to build features and technologies that will affect millions of your fellow developers in the community.", "spans": [{"start": 22, "end": 51, "label": "JOB_ROLE"}, {"start": 71, "end": 82, "label": "SKILL"}, {"start": 103, "end": 112, "label": "JOB_ROLE"}, {"start": 135, "end": 147, "label": "SKILL"}, {"start": 189, "end": 199, "label": "JOB_ROLE"}]}
    ]]

    tokenized_examples = [Example(**example) for example in [
        {"text": "Have you used the new version of my model?", "spans": [{"start": 36, "end": 41, "token_start": 8, "token_end": 8, "label": "SKILL"}], "tokens": [{"text": "Have", "start": 0, "end": 4, "id": 0}, {"text": "you", "start": 5, "end": 8, "id": 1}, {"text": "used", "start": 9, "end": 13, "id": 2}, {"text": "the", "start": 14, "end": 17, "id": 3}, {"text": "new", "start": 18, "end": 21, "id": 4}, {"text": "version", "start": 22, "end": 29, "id": 5}, {"text": "of", "start": 30, "end": 32, "id": 6}, {"text": "my", "start": 33, "end": 35, "id": 7}, {"text": "model", "start": 36, "end": 41, "id": 8}, {"text": "?", "start": 41, "end": 42, "id": 9}]},
        {"text": "I'd like to work as an actor or model if possible.", "spans": [{"text": "actor", "start": 23, "end": 28, "token_start": 7, "token_end": 7, "label": "JOB_ROLE"}, {"text": "model", "start": 32, "end": 37, "token_start": 9, "token_end": 9, "label": "JOB_ROLE"}], "tokens": [{"text": "I", "start": 0, "end": 1, "id": 0}, {"text": "'d", "start": 1, "end": 3, "id": 1}, {"text": "like", "start": 4, "end": 8, "id": 2}, {"text": "to", "start": 9, "end": 11, "id": 3}, {"text": "work", "start": 12, "end": 16, "id": 4}, {"text": "as", "start": 17, "end": 19, "id": 5}, {"text": "an", "start": 20, "end": 22, "id": 6}, {"text": "actor", "start": 23, "end": 28, "id": 7}, {"text": "or", "start": 29, "end": 31, "id": 8}, {"text": "model", "start": 32, "end": 37, "id": 9}, {"text": "if", "start": 38, "end": 40, "id": 10}, {"text": "possible", "start": 41, "end": 49, "id": 11}, {"text": ".", "start": 49, "end": 50, "id": 12}]},
        {"text": "We are looking for a Software Development Engineer who has solid coding skills, a strong machine learning background, and is passionate about developing new AI products.", "tokens": [{"text": "We", "start": 0, "end": 2, "id": 0}, {"text": "are", "start": 3, "end": 6, "id": 1}, {"text": "looking", "start": 7, "end": 14, "id": 2}, {"text": "for", "start": 15, "end": 18, "id": 3}, {"text": "a", "start": 19, "end": 20, "id": 4}, {"text": "Software", "start": 21, "end": 29, "id": 5}, {"text": "Development", "start": 30, "end": 41, "id": 6}, {"text": "Engineer", "start": 42, "end": 50, "id": 7}, {"text": "who", "start": 51, "end": 54, "id": 8}, {"text": "has", "start": 55, "end": 58, "id": 9}, {"text": "solid", "start": 59, "end": 64, "id": 10}, {"text": "coding", "start": 65, "end": 71, "id": 11}, {"text": "skills", "start": 72, "end": 78, "id": 12}, {"text": ",", "start": 78, "end": 79, "id": 13}, {"text": "a", "start": 80, "end": 81, "id": 14}, {"text": "strong", "start": 82, "end": 88, "id": 15}, {"text": "machine", "start": 89, "end": 96, "id": 16}, {"text": "learning", "start": 97, "end": 105, "id": 17}, {"text": "background", "start": 106, "end": 116, "id": 18}, {"text": ",", "start": 116, "end": 117, "id": 19}, {"text": "and", "start": 118, "end": 121, "id": 20}, {"text": "is", "start": 122, "end": 124, "id": 21}, {"text": "passionate", "start": 125, "end": 135, "id": 22}, {"text": "about", "start": 136, "end": 141, "id": 23}, {"text": "developing", "start": 142, "end": 152, "id": 24}, {"text": "new", "start": 153, "end": 156, "id": 25}, {"text": "AI", "start": 157, "end": 159, "id": 26}, {"text": "products", "start": 160, "end": 168, "id": 27}, {"text": ".", "start": 168, "end": 169, "id": 28}], "spans": [{"start": 21, "end": 50, "token_start": 5, "token_end": 7, "label": "SKILL"}, {"start": 65, "end": 71, "token_start": 11, "token_end": 11, "label": "SKILL"}, {"start": 89, "end": 105, "token_start": 16, "token_end": 17, "label": "SKILL"}, {"start": 142, "end": 152, "token_start": 24, "token_end": 24, "label": "SKILL"}, {"start": 157, "end": 159, "token_start": 26, "token_end": 26, "label": "SKILL"}]},
        {"text": "Responsibilities As a SOFTWARE DEVELOPMENT ENGINEER II you will work / collaborate with other talented engineers to build features and technologies that will affect millions of your fellow developers in the community.", "tokens": [{"text": "Responsibilities", "start": 0, "end": 16, "id": 0}, {"text": "As", "start": 17, "end": 19, "id": 1}, {"text": "a", "start": 20, "end": 21, "id": 2}, {"text": "SOFTWARE", "start": 22, "end": 30, "id": 3}, {"text": "DEVELOPMENT", "start": 31, "end": 42, "id": 4}, {"text": "ENGINEER", "start": 43, "end": 51, "id": 5}, {"text": "II", "start": 52, "end": 54, "id": 6}, {"text": "you", "start": 55, "end": 58, "id": 7}, {"text": "will", "start": 59, "end": 63, "id": 8}, {"text": "work", "start": 64, "end": 68, "id": 9}, {"text": "/", "start": 69, "end": 70, "id": 10}, {"text": "collaborate", "start": 71, "end": 82, "id": 11}, {"text": "with", "start": 83, "end": 87, "id": 12}, {"text": "other", "start": 88, "end": 93, "id": 13}, {"text": "talented", "start": 94, "end": 102, "id": 14}, {"text": "engineers", "start": 103, "end": 112, "id": 15}, {"text": "to", "start": 113, "end": 115, "id": 16}, {"text": "build", "start": 116, "end": 121, "id": 17}, {"text": "features", "start": 122, "end": 130, "id": 18}, {"text": "and", "start": 131, "end": 134, "id": 19}, {"text": "technologies", "start": 135, "end": 147, "id": 20}, {"text": "that", "start": 148, "end": 152, "id": 21}, {"text": "will", "start": 153, "end": 157, "id": 22}, {"text": "affect", "start": 158, "end": 164, "id": 23}, {"text": "millions", "start": 165, "end": 173, "id": 24}, {"text": "of", "start": 174, "end": 176, "id": 25}, {"text": "your", "start": 177, "end": 181, "id": 26}, {"text": "fellow", "start": 182, "end": 188, "id": 27}, {"text": "developers", "start": 189, "end": 199, "id": 28}, {"text": "in", "start": 200, "end": 202, "id": 29}, {"text": "the", "start": 203, "end": 206, "id": 30}, {"text": "community", "start": 207, "end": 216, "id": 31}, {"text": ".", "start": 216, "end": 217, "id": 32}], "spans": [{"start": 22, "end": 51, "token_start": 3, "token_end": 5, "label": "JOB_ROLE"}, {"start": 71, "end": 82, "token_start": 11, "token_end": 11, "label": "SKILL"}, {"start": 103, "end": 112, "token_start": 15, "token_end": 15, "label": "JOB_ROLE"}, {"start": 135, "end": 147, "token_start": 20, "token_end": 20, "label": "SKILL"}, {"start": 189, "end": 199, "token_start": 28, "token_end": 28, "label": "JOB_ROLE"}]}
    ]]
    # fmt: on

    fixed_examples = []
    for orig_example_hash, example, preprocessed_outputs in op_iter(
        untokenized_examples, pre=[spacy_preprocessor]
    ):
        fixed_examples.append(add_tokens(example, preprocessed_outputs=preprocessed_outputs))

    for fixed_example, tokenized_example in zip(fixed_examples, tokenized_examples):
        assert fixed_example.text == tokenized_example.text
        assert fixed_example.spans == tokenized_example.spans
        assert fixed_example.tokens == tokenized_example.tokens


def test_add_tokens_bad_example(spacy_preprocessor):

    bad_example = {
        "text": "The primary outcome was spontaneous delivery before 34 weeks.face=+Italic; Resultsface=-Italic; Spontaneous delivery before 34 weeks of gestation was less frequent in the progesterone group than in the placebo group (19.2% vs. 34.4%; relative risk, 0.56; 95% confidence interval [CI], 0.36 to 0.86).",
        "spans": [
            {"text": "34 weeks", "start": 61, "end": 69, "label": "DURATION"},
            {"text": "gestation", "start": 133, "end": 145, "label": "DURATION"},
            {"text": "19.2%", "start": 226, "end": 231, "label": "PERCENT"},
            {"text": "34.4%", "start": 236, "end": 241, "label": "PERCENT"},
            {"text": "0.56", "start": 258, "end": 262, "label": "NUMBER"},
            {"text": "95%", "start": 264, "end": 267, "label": "PERCENT"},
            {"text": "0.36 to 0.86", "start": 294, "end": 306, "label": "NUM_RANGE"},
        ],
    }

    examples = [Example(text=bad_example["text"], spans=[]), Example(**bad_example)]

    fixed_examples = []
    for orig_example_hash, example, preprocessed_outputs in op_iter(
        examples, pre=[spacy_preprocessor]
    ):
        fixed_example = add_tokens(example, preprocessed_outputs=preprocessed_outputs)
        if fixed_example:
            fixed_examples.append(fixed_example)

    assert len(fixed_examples) == 1
    fixed_example = fixed_examples[0]

    # Since add_tokens cannot resolve token start and ends from the spans above.
    assert isinstance(fixed_example, Example)

    assert isinstance(fixed_example.tokens, list)
    assert len(fixed_example.tokens) == 56



from recon.stats import (
    calculate_entity_coverage_entropy,
    calculate_entity_coverage_similarity,
    calculate_label_balance_entropy,
    calculate_label_distribution_similarity,
    detect_outliers,
    get_entity_coverage,
    get_ner_stats,
)
from recon.types import NERStats


def test_get_ner_stats(example_corpus_processed):
    stats: NERStats = get_ner_stats(example_corpus_processed.train)  # type: ignore
    str_stats = get_ner_stats(example_corpus_processed.train, serialize=True)
    assert isinstance(str_stats, str)

    assert stats.n_examples == 106
    assert stats.n_examples_no_entities == 29
    assert stats.n_annotations_per_type == {"SKILL": 199, "PRODUCT": 34, "JOB_ROLE": 10}


def test_calculate_label_distribution_similarity(example_corpus_processed):
    similarity = calculate_label_distribution_similarity(
        example_corpus_processed.train, example_corpus_processed.dev
    )

    assert round(similarity, 2) == 86.44


def test_calculate_entity_coverage_similarity(example_corpus_processed):
    ec_stats = calculate_entity_coverage_similarity(
        example_corpus_processed.train, example_corpus_processed.dev
    )

    assert round(ec_stats.entity, 2) == 31.21
    assert round(ec_stats.count, 2) == 36.67


def test_calculate_entity_coverage_entropy(example_corpus_processed):
    entity_coverage = get_entity_coverage(example_corpus_processed.train)
    entropy = calculate_entity_coverage_entropy(entity_coverage)

    assert round(entropy, 2) == 5.24


def test_calculate_label_balance_entropy(example_corpus_processed):
    ner_stats = get_ner_stats(example_corpus_processed.train)
    entropy = calculate_label_balance_entropy(ner_stats)  # type: ignore
    assert round(entropy, 2) == 0.57


def test_detect_outliers():
    seq = [-1, 10, 11, 12, 12, 13, 14, 15, 16, 16, 17, 99]

    outliers = detect_outliers(seq)
    assert outliers.low == [0]
    assert outliers.high == [11]



from pathlib import Path
from typing import Any, Dict, List, Set, Union

import srsly
from spacy.util import ensure_path

from .types import Example


class ExampleStore:
    def __init__(self, examples: List[Example] = None):
        self._map: Dict[int, Example] = {}
        if examples is not None:
            for e in examples:
                self.add(e)

    def __getitem__(self, example_hash: int) -> Example:
        return self._map[example_hash]

    def __len__(self) -> int:
        """The number of strings in the store.

        Returns:
            Number of examples in store
        """
        return len(self._map)

    def __contains__(self, example: Union[int, Example]) -> bool:
        """Check whether a string is in the store.
        
        Args:
            example (Union[int, Example]): The example to check

        Returns:
            Whether the store contains the example.
        """
        example_hash = hash(example) if isinstance(example, Example) else example
        return example_hash in self._map

    def add(self, example: Example) -> None:
        """Add an Example to the store
        
        Args:
            example (Example): example to add
        """
        example_hash = hash(example)
        self._map[example_hash] = example

    def from_disk(self, path: Path) -> "ExampleStore":
        """Load store from disk

        Args:
            path (Path): Path to file to load from
        
        Returns:
            ExampleStore: Initialized ExampleStore
        """
        path = ensure_path(path)
        examples = srsly.read_jsonl(path)
        for e in examples:
            example_hash = e["example_hash"]
            raw_example = e["example"]
            example = Example(**raw_example)
            assert hash(example) == example_hash
            self.add(example)

        return self

    def to_disk(self, path: Path) -> None:
        """Save store to disk
        
        Args:
            path (Path): Path to save store to
        """
        path = ensure_path(path)
        examples = []
        for example_hash, example in self._map.items():
            examples.append({"example_hash": example_hash, "example": example.dict()})

        srsly.write_jsonl(path, examples)



# isort:skip_file
# type: ignore
# -*- coding: utf-8 -*-
import typer
from pathlib import Path
from typing import Callable

from recon.corpus import Corpus
from recon.stats import (
    calculate_label_balance_entropy,
    calculate_label_distribution_similarity,
    get_ner_stats,
)
from recon.types import Example, NERStats
from wasabi import Printer
import dash
import dash_core_components as dcc
import dash_html_components as html
import plotly.graph_objs as go

from fastapi import FastAPI
import uvicorn

app = FastAPI()


@app.get("/")
def read_root():
    return {"Hello": "World"}


@app.get("/items/{item_id}")
def read_item(item_id: int, q: str = None):
    return {"item_id": item_id, "q": q}


def dashboard(data_dir: Path) -> None:
    """Calculate statistics on a Corpus

    Args:
        data_dir (Path): Path to data folder
    """
    msg: Printer = Printer()

    # with msg.loading("Loading Corpus from Disk"):
    corpus = Corpus.from_disk(data_dir)
    # msg.good("Done")

    ner_stats = corpus.apply(get_ner_stats)

    # external_stylesheets = [
    #     "https://codepen.io/chriddyp/pen/bWLwgP.css",
    #     "https://cdn.jsdelivr.net/npm/uikit@3.3.7/dist/css/uikit.min.css"
    # ]

    # external_scripts = [
    #     "https://cdn.jsdelivr.net/npm/uikit@3.3.7/dist/js/uikit.min.js",
    #     "https://cdn.jsdelivr.net/npm/uikit@3.3.7/dist/js/uikit-icons.min.js"
    # ]

    # app = dash.Dash(__name__, external_stylesheets=external_stylesheets, external_scripts=external_scripts)

    # def generate_bar_chart_stats(id: str, ner_stats: NERStats, name: str = None):
    #     return dcc.Graph(
    #         id=id,
    #         figure={
    #             'data': [
    #                 go.Bar(
    #                     x = list(ner_stats.n_annotations_per_type.values()),
    #                     y = list(ner_stats.n_annotations_per_type.keys()),
    #                     orientation='h'
    #                 )
    #             ],
    #             'layout': {
    #                 'title': name or id.capitalize()
    #             }
    #         }
    #     )

    # app.layout = html.Div(children=[
    #     html.Div(className="uk-child-width-1-2@s uk-grid-match")
    #     html.H1(className="" children='Recon NER Dashboard'),

    #     html.Div(children='''
    #         This dashboard shows statistics for all your data.
    #     '''),

    #     html.Div(children=[
    #         generate_bar_chart_stats("train", ner_stats["train"]),
    #         generate_bar_chart_stats("dev", ner_stats["dev"]),
    #         generate_bar_chart_stats("test", ner_stats["test"]),
    #         generate_bar_chart_stats("all", ner_stats["all"])
    #     ], style={'columnCount': 4})
    # ])

    uvicorn.run(app, port=9090)

    # app.run_server(debug=True)


if __name__ == "__main__":
    typer.run(dashboard)



import functools
from collections import Counter, defaultdict
from copy import deepcopy
from inspect import isclass
from typing import Any, Callable, Dict, Iterator, List, Set, Tuple, Union

import catalogue
import srsly
from wasabi import Printer

from .preprocess import PreProcessor
from .types import (
    Example,
    OperationResult,
    OperationState,
    OperationStatus,
    Transformation,
    TransformationCallbacks,
    TransformationType,
)


class registry:
    operations = catalogue.create("recon", "operations", entry_points=True)


def op_iter(
    data: List[Example], pre: List[PreProcessor], verbose: bool = True
) -> Iterator[Tuple[int, Example, Dict[str, Any]]]:
    """Iterate over list of examples for an operation
    yielding tuples of (example hash, example)
    
    Args:
        data (List[Example]): List of examples to iterate
        pre (List[PreProcessor]): List of preprocessors to run
        verbose (bool, optional): Show verbose output.
    
    Yields:
        Iterator[Tuple[int, Example]]: Tuples of (example hash, example)
    """
    msg = Printer(no_print=verbose == False, hide_animation=verbose == False)
    preprocessed_outputs: Dict[Example, Dict[str, Any]] = defaultdict(dict)
    for processor in pre:
        with msg.loading(f"\t=> Running preprocessor {processor.name}..."):
            processor_outputs = list(processor(data))
            msg.good("Done")

        for i, (example, output) in enumerate(zip(data, processor_outputs)):
            preprocessed_outputs[example][processor.name] = processor_outputs[i]

    for example in data:
        yield hash(example), example.copy(deep=True), preprocessed_outputs[example]


class operation:
    def __init__(self, name: str, pre: List[PreProcessor] = []):
        """Decorate an operation that makes some changes to a dataset.

        Args:
            name (str): Operation name.
            pre (List[PreProcessor]): List of preprocessors to run
        """
        self.name = name
        self.pre = pre

    def __call__(self, *args: Any, **kwargs: Any) -> Callable:
        """Decorator for an operation. 
        The first arg is the function being decorated.
        This function can either operate on a List[Example]
        and in that case self.batch should be True.

        e.g. @operation("recon.v1.some_name", batch=True)

        Or it should operate on a single example and 
        recon will take care of applying it to a full Dataset
        
        Args:
            args: First arg is function to decorate
        
        Returns:
            Callable: Original function
        """
        op: Callable = args[0]
        registry.operations.register(self.name)(Operation(self.name, self.pre, op))

        return op


class Operation:
    """Operation class that takes care of calling and reporting
    the results of an operation on a Dataset"""

    def __init__(self, name: str, pre: List[PreProcessor], op: Callable):
        """Initialize an Operation instance
        
        Args:
            name (str): Name of operation
            pre (List[PreProcessor]): List of preprocessors to run
            op (Callable): Decorated function
        """
        self.name = name
        self.pre = pre
        self.op = op

    def __call__(self, dataset: Any, *args: Any, **kwargs: Any) -> OperationResult:
        """Runs op on a dataset and records the results
        
        Args:
            dataset (Dataset): Dataset to operate on
        
        Raises:
            ValueError: if track_example is called in the op with no data
        
        Returns:
            OperationResult: Container holding new data and the state of the Operation
        """
        verbose = True
        msg = Printer(no_print=verbose == False)

        initial_state = kwargs.pop("initial_state") if "initial_state" in kwargs else None
        verbose = kwargs.pop("verbose") if "verbose" in kwargs else None
        if not initial_state:
            initial_state = OperationState(name=self.name)
        state = initial_state.copy(deep=True)

        if state.status == OperationStatus.NOT_STARTED:
            state.status = OperationStatus.IN_PROGRESS

        def add_example(new_example: Example) -> None:
            state.transformations.append(
                Transformation(example=hash(new_example), type=TransformationType.EXAMPLE_ADDED)
            )
            dataset.example_store.add(new_example)

        def remove_example(orig_example_hash: int) -> None:
            state.transformations.append(
                Transformation(
                    prev_example=orig_example_hash, type=TransformationType.EXAMPLE_REMOVED
                )
            )

        def change_example(orig_example_hash: int, new_example: Example) -> None:
            state.transformations.append(
                Transformation(
                    prev_example=orig_example_hash,
                    example=hash(new_example),
                    type=TransformationType.EXAMPLE_CHANGED,
                )
            )
            dataset.example_store.add(new_example)

        new_data = []
        for orig_example_hash, example, preprocessed_outputs in op_iter(
            dataset.data, self.pre, verbose=verbose
        ):
            if preprocessed_outputs:
                res = self.op(example, *args, preprocessed_outputs=preprocessed_outputs, **kwargs)
            else:
                res = self.op(example, *args, **kwargs)

            if res is None:
                remove_example(orig_example_hash)
            elif isinstance(res, list):
                old_example_present = False
                for new_example in res:
                    if hash(new_example) == orig_example_hash:
                        old_example_present = True
                    else:
                        new_data.append(new_example)
                        add_example(new_example)
                if not old_example_present:
                    remove_example(orig_example_hash)
            else:
                assert isinstance(res.text, str)
                assert isinstance(res.spans, list)
                new_data.append(res)
                if hash(res) != orig_example_hash:
                    change_example(orig_example_hash, res)

        transformation_counts = Counter([t.type for t in state.transformations])

        state.examples_added = transformation_counts[TransformationType.EXAMPLE_ADDED]
        state.examples_removed = transformation_counts[TransformationType.EXAMPLE_REMOVED]
        state.examples_changed = transformation_counts[TransformationType.EXAMPLE_CHANGED]
        state.status = OperationStatus.COMPLETED

        state_copy = state.copy(deep=True)
        state = OperationState(name=self.name)
        return OperationResult(data=new_data, state=state_copy)



from typing import Dict, Iterable, Iterator, List, Set, Tuple

import srsly
from spacy.language import Language
from wasabi import Printer

from .types import Example, Span, Token


class EntityRecognizer:
    """Abstract Base class for recognizing entities in a batch of text.
    Used in the `recon.insights` module for understanding the kinds
    of examples your model is having the most trouble with.
    """

    @property
    def labels(self) -> List[str]:
        """Return List of String Labels
        
        Raises:
            NotImplementedError: Not Implemented, override
        
        Returns:
            List[str]: List of labels the model can predict
        """
        raise NotImplementedError

    def predict(self, texts: Iterable[str]) -> Iterator[Example]:
        """Run model inference on a batch of raw texts.
        
        Args:
            texts (Iterable[str]): Raw text examples
        
        Raises:
            NotImplementedError: Not implemented, override
        
        Returns:
            Iterator[Example]: Iterator of Examples
        """
        raise NotImplementedError


class SpacyEntityRecognizer(EntityRecognizer):
    """Create an EntityRecognizer from a spaCy Langauge instance"""

    def __init__(self, nlp: Language):
        """Initialize a SpacyEntityRecognizer
        
        Args:
            nlp (Language): spaCy Language instance that can sets doc.ents
        """
        super().__init__()
        self.nlp = nlp

    @property
    def labels(self) -> List[str]:
        """Return List of spaCy ner labels
        
        Returns:
            List[str]: List of labels from spaCy ner pipe
        """
        all_labels: Set[str] = set()

        for pipe in ["ner", "entity_ruler"]:
            if self.nlp.has_pipe(pipe):
                all_labels = all_labels | set(self.nlp.get_pipe(pipe).labels)

        return sorted(list(all_labels))

    def predict(self, texts: Iterable[str]) -> Iterator[Example]:
        """Run spaCy nlp.pipe on a batch of raw texts.
        
        Args:
            texts (Iterable[str]): Raw text examples
        
        Yields:
            Iterator[Example]: Examples constructed from spaCy Model predictions
        """
        examples: List[Example] = []

        for doc in self.nlp.pipe(texts):
            yield Example(
                text=doc.text,
                spans=[
                    Span(
                        text=e.text,
                        start=e.start_char,
                        end=e.end_char,
                        label=e.label_,
                        token_start=e.start,
                        token_end=e.end,
                    )
                    for e in doc.ents
                ],
                tokens=[Token(text=t.text, start=t.idx, end=t.idx + len(t), id=t.i) for t in doc],
            )

    def evaluate(self, data: List[Example]) -> None:
        msg = Printer()
        formatted_data, _ = self._format_data(data)
        sc = self.nlp.evaluate(formatted_data, batch_size=64)
        msg.divider("Recognizer Results")
        result = [
            ("Precision", f"{sc.ents_p:.3f}"),
            ("Recall", f"{sc.ents_r:.3f}"),
            ("F-Score", f"{sc.ents_f:.3f}"),
        ]
        msg.table(result)

        table_data = []
        for label, scores in sorted(sc.ents_per_type.items(), key=lambda tup: tup[0]):
            table_data.append(
                (label, f"{scores['p']:.3f}", f"{scores['r']:.3f}", f"{scores['f']:.3f}")
            )
        header = ("Label", "Precision", "Recall", "F-Score")
        formatted = msg.table(table_data, header=header, divider=True)
        return sc

    def _format_data(
        self, data: List[Example]
    ) -> Tuple[List[Tuple[str, Dict[str, List[Tuple[int, int, str]]]]], Set[str]]:
        result = []
        labels = set()
        for example in data:
            ents = [(s.start, s.end, s.label) for s in example.spans]
            labels.update([ent[2] for ent in ents])
            result.append((example.text, {"entities": ents}))
        return result, labels



from typing import cast

import pytest
from recon.stats import get_ner_stats
from recon.types import Example, NERStats, Span, Token
from recon.validation import filter_overlaps, upcase_labels


@pytest.fixture()
def messy_data():
    return [
        {
            "text": "Denver, Colorado is a city.",
            "spans": [
                {"start": 0, "end": 6, "label": "GPE"},
                {"start": 0, "end": 16, "label": "LOC"},
            ],
            "meta": "Cities Data",
        }
    ]


def test_upcase_labels(example_corpus):
    stats = cast(NERStats, get_ner_stats(example_corpus.train))
    assert "skill" in stats.n_annotations_per_type
    assert "product" in stats.n_annotations_per_type
    assert "SKILL" in stats.n_annotations_per_type

    example_corpus._train.apply_("recon.v1.upcase_labels")
    fixed_stats = cast(NERStats, get_ner_stats(example_corpus.train))
    assert "skill" not in fixed_stats.n_annotations_per_type
    assert "product" not in fixed_stats.n_annotations_per_type


def test_filter_overlaps():
    def get_test_example(span_offsets):
        spans = []
        for so in span_offsets:
            spans.append(Span(text="x" * (so[1] - so[0]), start=so[0], end=so[1], label=so[2]))
        return Example(text="x" * 1500, spans=spans)

    def spans_to_offsets(spans):
        return [(span.start, span.end, span.label) for span in spans]

    test_entities = [(0, 5, "ENTITY"), (6, 10, "ENTITY")]
    result = filter_overlaps(get_test_example(test_entities))
    assert spans_to_offsets(result.spans) == [(0, 5, "ENTITY"), (6, 10, "ENTITY")]

    test_entities = [(0, 5, "ENTITY"), (5, 10, "ENTITY")]
    result = filter_overlaps(get_test_example(test_entities))
    assert spans_to_offsets(result.spans) == [(0, 5, "ENTITY"), (5, 10, "ENTITY")]

    test_entities = [(0, 5, "ENTITY"), (4, 10, "ENTITY")]
    result = filter_overlaps(get_test_example(test_entities))
    assert spans_to_offsets(result.spans) == [(4, 10, "ENTITY")]

    test_entities = [(0, 5, "ENTITY"), (0, 5, "ENTITY")]
    result = filter_overlaps(get_test_example(test_entities))
    assert spans_to_offsets(result.spans) == [(0, 5, "ENTITY")]

    test_entities = [(0, 5, "ENTITY"), (4, 11, "ENTITY"), (6, 20, "ENTITY")]
    result = filter_overlaps(get_test_example(test_entities))
    assert spans_to_offsets(result.spans) == [(0, 5, "ENTITY"), (6, 20, "ENTITY")]

    test_entities = [(0, 5, "ENTITY"), (4, 7, "ENTITY"), (10, 20, "ENTITY")]
    result = filter_overlaps(get_test_example(test_entities))
    assert spans_to_offsets(result.spans) == [(0, 5, "ENTITY"), (10, 20, "ENTITY")]

    test_entities = [(1368, 1374, "ENTITY"), (1368, 1376, "ENTITY")]
    result = filter_overlaps(get_test_example(test_entities))
    assert spans_to_offsets(result.spans) == [(1368, 1376, "ENTITY")]

    test_entities = [
        (12, 33, "ENTITY"),
        (769, 779, "ENTITY"),
        (769, 787, "ENTITY"),
        (806, 811, "ENTITY"),
    ]
    result = filter_overlaps(get_test_example(test_entities))
    assert spans_to_offsets(result.spans) == [
        (12, 33, "ENTITY"),
        (769, 787, "ENTITY"),
        (806, 811, "ENTITY"),
    ]

    test_entities = [
        (189, 209, "ENTITY"),
        (317, 362, "ENTITY"),
        (345, 354, "ENTITY"),
        (364, 368, "ENTITY"),
    ]
    result = filter_overlaps(get_test_example(test_entities))
    assert spans_to_offsets(result.spans) == [
        (189, 209, "ENTITY"),
        (317, 362, "ENTITY"),
        (364, 368, "ENTITY"),
    ]

    test_entities = [(445, 502, "ENTITY"), (461, 473, "ENTITY"), (474, 489, "ENTITY")]
    result = filter_overlaps(get_test_example(test_entities))
    assert spans_to_offsets(result.spans) == [(445, 502, "ENTITY")]



from typing import Dict


class Scorer:
    def __init__(self) -> None:
        self._scores: Dict[str, object] = {}

    @property
    def scores(self) -> Dict[str, object]:
        return self._scores



"""ReconNER, Debug annotated Named Entity Recognition (NER) data for inconsitencies and get insights on improving the quality of your data."""

__version__ = "0.5.1"

from .corpus import Corpus
from .corrections import *
from .insights import *
from .loaders import read_json, read_jsonl
from .operations import operation
from .stats import get_ner_stats
from .tokenization import add_tokens, fix_tokenization_and_spacing
from .validation import filter_overlaps, upcase_labels

try:
    # This needs to be imported in order for the entry points to be loaded
    from .prodigy import recipes as prodigy_recipes  # noqa: F401
except ImportError:
    pass



from typing import Any, Callable, Dict, List

import catalogue
import spacy
from spacy.language import Language

from .types import Example

loading_pipelines = catalogue.create("recon", "loading_pipelines", entry_points=True)
tokenizers = catalogue.create("recon", "tokenizers", entry_points=True)


@loading_pipelines.register("default")
def default_loading_pipeline() -> List[str]:
    return [
        "fix_tokenization_and_spacing",
        "add_tokens",
        "fix_annotations_format",
        "filter_overlaps",
        "json_to_examples",
    ]


@tokenizers.register("default")
def default_tokenizer() -> Language:
    return spacy.blank("en")



import copy
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, List, Set, Tuple, Union, cast

import srsly
from spacy.util import ensure_path
from wasabi import Printer

from .hashing import dataset_hash
from .loaders import read_json, read_jsonl
from .operations import registry
from .registry import loading_pipelines
from .store import ExampleStore
from .types import (
    DatasetOperationsState,
    Example,
    OperationResult,
    OperationState,
    OperationStatus,
)


class Dataset:
    """A Dataset is a around a List of examples.
    Datasets are responsible for tracking all Operations done on them.
    This ensures data lineage and easy reporting of how changes in the 
    data based on various Operations effects overall quality.
    
    Dataset holds state (let's call it self.operations for now)
    self.operations is a list of every function run on the Dataset since it's
    initial creation. If loading from disk, track everything that happens in loading
    phase in operations as well by simply initializing self.operations in constructors
    
    Each operation should has the following attributes:
        operation hash
        name: function/callable name ideally, could be added with a decorator
        status: (not_started|completed)
        transformations: List[Transformation]
            commit hash
            timestamp(s) - start and end both? end is probably enough
            examples deleted
            examples added
            examples corrected
            annotations deleted
            annotations added
            annotations corrected

            for annotations deleted/added/corrected, include mapping from old Example hash to new Example hash
            that can be decoded for display later

    All operations are serializable in the to_disk and from_disk methods.

    So if I have 10 possible transformations.

    I can run 1..5, save to disk train a model and check results. 
    Then I can load that model from disk with all previous operations already tracked
    in self.operations. Then I can run 6..10, save to disk and train model.
    Now I have git-like "commits" for the data used in each model.
    
    """

    def __init__(
        self,
        name: str,
        data: List[Example] = [],
        operations: List[OperationState] = None,
        example_store: ExampleStore = None,
        verbose: bool = False,
    ):
        self.name = name
        self.data = data
        if not operations:
            operations = []
        self.operations = operations

        if example_store is None:
            example_store = ExampleStore(data)
        self.example_store = example_store
        self.verbose = verbose

    @property
    def commit_hash(self) -> str:
        return cast(str, dataset_hash(self, as_int=False))

    def __hash__(self) -> int:
        return cast(int, dataset_hash(self))

    def __len__(self) -> int:
        return len(self.data)

    def apply(
        self, func: Callable[[List[Example], Any, Any], Any], *args: Any, **kwargs: Any
    ) -> Any:
        """Apply a function to the dataset
        
        Args:
            func (Callable[[List[Example], Any, Any], Any]): 
                Function from an existing recon module that can operate on a List of examples
        
        Returns:
            Result of running func on List of examples
        """
        return func(self.data, *args, **kwargs)  # type: ignore

    def apply_(
        self,
        operation: Union[str, Callable[[Any], OperationResult]],
        *args: Any,
        initial_state: OperationState = None,
        **kwargs: Any,
    ) -> None:
        """Apply an operation to all data inplace.
        
        Args:
            operation (Callable[[Any], OperationResult]): Any operation that
                changes data in place. See recon.operations.registry.operations
        """
        if isinstance(operation, str):
            operation = registry.operations.get(operation)
            if operation:
                operation = cast(Callable, operation)

        name = getattr(operation, "name", None)
        if name is None or name not in registry.operations:
            raise ValueError(
                "This function is not an operation. Ensure your function is registered in the operations registry."
            )

        msg = Printer(no_print=self.verbose == False)
        msg.text(f"=> Applying operation '{name}' inplace")
        result: OperationResult = operation(self, *args, initial_state=initial_state, verbose=self.verbose, **kwargs)  # type: ignore
        msg.good(f"Completed operation '{name}'")

        self.operations.append(result.state)
        dataset_changed = any(
            (
                result.state.examples_added,
                result.state.examples_removed,
                result.state.examples_changed,
            )
        )
        if dataset_changed:
            self.data = result.data

    def pipe_(self, operations: List[Union[str, OperationState]]) -> None:
        """Run a sequence of operations on dataset data.
        Internally calls Dataset.apply_ and will resolve named
        operations in registry.operations
        
        Args:
            operations (List[Union[str, OperationState]]): List of operations
        """

        msg = Printer(no_print=self.verbose == False)
        msg.text(f"Applying pipeline of operations inplace to the dataset: {self.name}")

        for op in operations:
            op_name = op.name if isinstance(op, OperationState) else op
            msg.text(f"|_ {op_name}")

        for op in operations:
            if isinstance(op, str):
                op_name = op
                args = []
                kwargs = {}
                initial_state = None
            elif isinstance(op, OperationState):
                op_name = op.name
                args = op.args
                kwargs = op.kwargs
                initial_state = op

            operation = registry.operations.get(op_name)

            self.apply_(operation, *args, initial_state=initial_state, **kwargs)

    def from_disk(self, path: Path, loader_func: Callable = read_jsonl) -> "Dataset":
        """Load Dataset from disk given a path and a loader function that reads the data
        and returns an iterator of Examples
        
        Args:
            path (Path): path to load from
            loader_func (Callable, optional): Callable that reads a file and returns a List of examples. 
                Defaults to [read_jsonl][recon.loaders.read_jsonl]
        """
        path = ensure_path(path)
        ds_op_state = None
        if (path.parent / ".recon" / self.name).exists():
            state = srsly.read_json(path.parent / ".recon" / self.name / "state.json")
            ds_op_state = DatasetOperationsState(**state)
            self.operations = ds_op_state.operations

        data = loader_func(path)
        self.data = data
        for example in self.data:
            self.example_store.add(example)

        if ds_op_state and self.commit_hash != ds_op_state.commit:
            # Dataset changed, examples added
            self.operations.append(
                OperationState(
                    name="examples_added_external",
                    status=OperationStatus.COMPLETED,
                    ts=datetime.now(),
                    examples_added=max(len(self) - ds_op_state.size, 0),
                    examples_removed=max(ds_op_state.size - len(self), 0),
                    examples_changed=0,
                    transformations=[],
                )
            )

            for op in self.operations:
                op.status = OperationStatus.NOT_STARTED

        seen: Set[str] = set()
        operations_to_run: Dict[str, OperationState] = {}

        for op in self.operations:
            if (
                op.name not in operations_to_run
                and op.name in registry.operations
                and op.status != OperationStatus.COMPLETED
            ):
                operations_to_run[op.name] = op

        for op_name, state in operations_to_run.items():
            op = registry.operations.get(op_name)
            self.apply_(op, *state.args, initial_state=state, **state.kwargs)  # type: ignore

        return self

    def to_disk(self, output_path: Path, force: bool = False, save_examples: bool = True) -> None:
        """Save Corpus to Disk
        
        Args:
            output_path (Path): Output file path to save data to
            force (bool): Force save to directory. Create parent directories
                or overwrite existing data.
            save_examples (bool): Save the example store along with the state.
        """
        output_path = ensure_path(output_path)
        output_dir = output_path.parent
        state_dir = output_dir / ".recon" / self.name
        if force:
            output_dir.mkdir(parents=True, exist_ok=True)

            if not state_dir.exists():
                state_dir.mkdir(parents=True, exist_ok=True)

        ds_op_state = DatasetOperationsState(
            name=self.name, commit=self.commit_hash, size=len(self), operations=self.operations
        )
        srsly.write_json(state_dir / "state.json", ds_op_state.dict())

        if save_examples:
            self.example_store.to_disk(state_dir / "example_store.jsonl")

        srsly.write_jsonl(output_path, [e.dict() for e in self.data])



from pathlib import Path
from typing import Any, Callable, Dict, List, Union

import srsly
from spacy.util import ensure_path

from .dataset import Dataset
from .loaders import read_json, read_jsonl
from .store import ExampleStore
from .types import CorpusApplyResult, Example, OperationResult, OperationState


class Corpus:
    """Container for a full Corpus with train/dev/test splits.
    Used to apply core functions to all datasets at once.
    """

    def __init__(
        self, train: Dataset, dev: Dataset, test: Dataset = None, example_store: ExampleStore = None
    ):
        """Initialize a Corpus.
        
        Args:
            train (Dataset): List of examples for **train** set
            dev (Dataset): List of examples for **dev** set
            test (Dataset, optional): Defaults to None. List of examples for **test** set
        """
        if example_store is None:
            examples = train.data + dev.data
            if test:
                examples += test.data
            example_store = ExampleStore(examples)
        self.example_store = example_store

        if test is None:
            test = Dataset("test")

        for ds in (train, dev, test):
            ds.example_store = example_store

        self._train = train
        self._dev = dev
        self._test = test

    @property
    def train(self) -> List[Example]:
        """Return train dataset
        
        Returns:
            List[Example]: Train Examples
        """
        return self._train.data

    @property
    def dev(self) -> List[Example]:
        """Return train dev
        
        Returns:
            List[Example]: Train Examples
        """
        return self._dev.data

    @property
    def test(self) -> List[Example]:
        """Return test dataset
        
        Returns:
            List[Example]: Test Examples
        """
        return self._test.data or []

    @property
    def all(self) -> List[Example]:
        """Return concatenation of train/dev/test datasets
        
        Returns:
            List[Example]: All Examples in Corpus
        """
        return self.train + self.dev + self.test

    def apply(
        self, func: Callable[[List[Example], Any, Any], Any], *args: Any, **kwargs: Any
    ) -> CorpusApplyResult:
        """Apply a function to all datasets
        
        Args:
            func (Callable[[List[Example], Any, Any], Any]): 
                Function from an existing recon module that can operate on a List of examples
        
        Returns:
            CorpusApplyResult: CorpusApplyResult mapping dataset name to return type of func Callable
        """

        return CorpusApplyResult(
            train=func(self.train, *args, **kwargs),  # type: ignore
            dev=func(self.dev, *args, **kwargs),  # type: ignore
            test=func(self.test, *args, **kwargs),  # type: ignore
            all=func(self.all, *args, **kwargs),  # type: ignore
        )

    def apply_(
        self, operation: Callable[[Any], OperationResult], *args: Any, **kwargs: Any
    ) -> None:
        """Apply a function to all data inplace.
        
        Args:
            operation (Callable[[Any], OperationResult]): Any operation that
                changes data in place. See recon.operations.registry.operations
        """
        self._train.apply_(operation, *args, **kwargs)
        self._dev.apply_(operation, *args, **kwargs)
        self._test.apply_(operation, *args, **kwargs)

    def pipe_(self, operations: List[Union[str, OperationState]]) -> None:
        """Run a sequence of operations on each dataset.
        Calls Dataset.pipe_ for each dataset
        
        Args:
            operations (List[Union[str, OperationState]]): List of operations
        """
        self._train.pipe_(operations)
        self._dev.pipe_(operations)
        self._test.pipe_(operations)

    @classmethod
    def from_disk(
        cls,
        data_dir: Path,
        train_file: str = "train.jsonl",
        dev_file: str = "dev.jsonl",
        test_file: str = "test.jsonl",
        loader_func: Callable = read_jsonl,
    ) -> "Corpus":
        """Load Corpus from disk given a directory with files 
        named explicitly train.jsonl, dev.jsonl, and test.jsonl
        
        Args:
            data_dir (Path): directory to load from.
            train_file (str, optional): Filename of train data under data_dir. Defaults to train.jsonl.
            dev_file (str, optional): Filename of dev data under data_dir. Defaults to dev.jsonl.
            test_file (str, optional): Filename of test data under data_dir. Defaults to test.jsonl.
            loader_func (Callable, optional): Callable that reads a file and returns a List of examples. 
                Defaults to [read_jsonl][recon.loaders.read_jsonl]
        """
        data_dir = ensure_path(data_dir)

        train = Dataset("train").from_disk(data_dir / train_file)
        dev = Dataset("dev").from_disk(data_dir / dev_file)

        try:
            test = Dataset("test").from_disk(data_dir / test_file)
            corpus = cls(train, dev, test=test)
        except ValueError as e:
            corpus = cls(train, dev)
        return corpus

    def to_disk(self, data_dir: Path, force: bool = False) -> None:
        """Save Corpus to Disk
        
        Args:
            data_dir (Path): Directory to save data to
            force (bool): Force save to directory. Create parent directories
                or overwrite existing data.
        """
        data_dir = ensure_path(data_dir)
        state_dir = data_dir / ".recon"
        if force:
            data_dir.mkdir(parents=True, exist_ok=True)

            if not state_dir.exists():
                state_dir.mkdir(parents=True, exist_ok=True)

        self._train.to_disk(data_dir / "train.jsonl", force=force, save_examples=False)
        self._dev.to_disk(data_dir / "dev.jsonl", force=force, save_examples=False)
        if self._test:
            self._test.to_disk(data_dir / "test.jsonl", force=force, save_examples=False)

        self.example_store.to_disk(state_dir / "example_store.jsonl")



from pathlib import Path

import typer
from recon.dataset import Dataset
from recon.stats import get_ner_stats


def main(data_file: Path, output_file: Path):
    ds = Dataset("train").from_disk(data_file)

    print("STATS BEFORE")
    print("============")
    print(ds.apply(get_ner_stats, serialize=True))

    ds.apply_("recon.v1.upcase_labels")

    print("STATS AFTER")
    print("===========")
    print(ds.apply(get_ner_stats, serialize=True))


if __name__ == "__main__":
    typer.run(main)



import pytest
from recon.corrections import corrections_from_dict, fix_annotations, rename_labels
from recon.insights import get_label_disparities
from recon.types import Example


@pytest.fixture()
def test_examples():
    # fmt: off
    raw_examples = [
        {"text":"Have you used the new version of my model?","spans":[{"start":36,"end":41,"token_start":8,"token_end":8,"label":"SKILL"}],"_input_hash":1798863398,"_task_hash":1273875979,"tokens":[{"text":"Have","start":0,"end":4,"id":0},{"text":"you","start":5,"end":8,"id":1},{"text":"used","start":9,"end":13,"id":2},{"text":"the","start":14,"end":17,"id":3},{"text":"new","start":18,"end":21,"id":4},{"text":"version","start":22,"end":29,"id":5},{"text":"of","start":30,"end":32,"id":6},{"text":"my","start":33,"end":35,"id":7},{"text":"model","start":36,"end":41,"id":8},{"text":"?","start":41,"end":42,"id":9}],"_session_id":None,"_view_id":"ner_manual","answer":"accept"},
        {"text":"I'd like to work as an actor or model if possible.","spans":[{"text":"actor","start":23,"end":28,"token_start":7,"token_end":7,"label":"JOB_ROLE"},{"text":"model","start":32,"end":37,"token_start":9,"token_end":9,"label":"JOB_ROLE"}],"_input_hash":895552771,"_task_hash":-936257555,"tokens":[{"text":"I","start":0,"end":1,"id":0},{"text":"'d","start":1,"end":3,"id":1},{"text":"like","start":4,"end":8,"id":2},{"text":"to","start":9,"end":11,"id":3},{"text":"work","start":12,"end":16,"id":4},{"text":"as","start":17,"end":19,"id":5},{"text":"an","start":20,"end":22,"id":6},{"text":"actor","start":23,"end":28,"id":7},{"text":"or","start":29,"end":31,"id":8},{"text":"model","start":32,"end":37,"id":9},{"text":"if","start":38,"end":40,"id":10},{"text":"possible","start":41,"end":49,"id":11},{"text":".","start":49,"end":50,"id":12}],"_session_id":None,"_view_id":"ner_manual","answer":"accept"},
        {"text":"We are looking for a Software Development Engineer who has solid coding skills, a strong machine learning background, and is passionate about developing new AI products.","meta":{"id":"599073","category":"Engineering","subCategory":"Data & Applied Sciences"},"_input_hash":-275722772,"_task_hash":868828486,"_session_id":None,"_view_id":"ner_manual","answer":"accept","dataset":1,"tokens":[{"text":"We","start":0,"end":2,"id":0},{"text":"are","start":3,"end":6,"id":1},{"text":"looking","start":7,"end":14,"id":2},{"text":"for","start":15,"end":18,"id":3},{"text":"a","start":19,"end":20,"id":4},{"text":"Software","start":21,"end":29,"id":5},{"text":"Development","start":30,"end":41,"id":6},{"text":"Engineer","start":42,"end":50,"id":7},{"text":"who","start":51,"end":54,"id":8},{"text":"has","start":55,"end":58,"id":9},{"text":"solid","start":59,"end":64,"id":10},{"text":"coding","start":65,"end":71,"id":11},{"text":"skills","start":72,"end":78,"id":12},{"text":",","start":78,"end":79,"id":13},{"text":"a","start":80,"end":81,"id":14},{"text":"strong","start":82,"end":88,"id":15},{"text":"machine","start":89,"end":96,"id":16},{"text":"learning","start":97,"end":105,"id":17},{"text":"background","start":106,"end":116,"id":18},{"text":",","start":116,"end":117,"id":19},{"text":"and","start":118,"end":121,"id":20},{"text":"is","start":122,"end":124,"id":21},{"text":"passionate","start":125,"end":135,"id":22},{"text":"about","start":136,"end":141,"id":23},{"text":"developing","start":142,"end":152,"id":24},{"text":"new","start":153,"end":156,"id":25},{"text":"AI","start":157,"end":159,"id":26},{"text":"products","start":160,"end":168,"id":27},{"text":".","start":168,"end":169,"id":28}],"spans":[{"start":21,"end":50,"token_start":5,"token_end":7,"label":"SKILL"},{"start":65,"end":71,"token_start":11,"token_end":11,"label":"SKILL"},{"start":89,"end":105,"token_start":16,"token_end":17,"label":"SKILL"},{"start":142,"end":152,"token_start":24,"token_end":24,"label":"SKILL"},{"start":157,"end":159,"token_start":26,"token_end":26,"label":"SKILL"}]},
        {"text":"Responsibilities As a SOFTWARE DEVELOPMENT ENGINEER II you will work / collaborate with other talented engineers to build features and technologies that will affect millions of your fellow developers in the community.","meta":{"id":"598569","category":"Engineering","subCategory":"Software Engineering"},"_input_hash":-930202924,"_task_hash":547318084,"_session_id":None,"_view_id":"ner_manual","answer":"accept","dataset":1,"tokens":[{"text":"Responsibilities","start":0,"end":16,"id":0},{"text":"As","start":17,"end":19,"id":1},{"text":"a","start":20,"end":21,"id":2},{"text":"SOFTWARE","start":22,"end":30,"id":3},{"text":"DEVELOPMENT","start":31,"end":42,"id":4},{"text":"ENGINEER","start":43,"end":51,"id":5},{"text":"II","start":52,"end":54,"id":6},{"text":"you","start":55,"end":58,"id":7},{"text":"will","start":59,"end":63,"id":8},{"text":"work","start":64,"end":68,"id":9},{"text":"/","start":69,"end":70,"id":10},{"text":"collaborate","start":71,"end":82,"id":11},{"text":"with","start":83,"end":87,"id":12},{"text":"other","start":88,"end":93,"id":13},{"text":"talented","start":94,"end":102,"id":14},{"text":"engineers","start":103,"end":112,"id":15},{"text":"to","start":113,"end":115,"id":16},{"text":"build","start":116,"end":121,"id":17},{"text":"features","start":122,"end":130,"id":18},{"text":"and","start":131,"end":134,"id":19},{"text":"technologies","start":135,"end":147,"id":20},{"text":"that","start":148,"end":152,"id":21},{"text":"will","start":153,"end":157,"id":22},{"text":"affect","start":158,"end":164,"id":23},{"text":"millions","start":165,"end":173,"id":24},{"text":"of","start":174,"end":176,"id":25},{"text":"your","start":177,"end":181,"id":26},{"text":"fellow","start":182,"end":188,"id":27},{"text":"developers","start":189,"end":199,"id":28},{"text":"in","start":200,"end":202,"id":29},{"text":"the","start":203,"end":206,"id":30},{"text":"community","start":207,"end":216,"id":31},{"text":".","start":216,"end":217,"id":32}],"spans":[{"start":22,"end":51,"token_start":3,"token_end":5,"label":"JOB_ROLE"},{"start":71,"end":82,"token_start":11,"token_end":11,"label":"SKILL"},{"start":103,"end":112,"token_start":15,"token_end":15,"label":"JOB_ROLE"},{"start":135,"end":147,"token_start":20,"token_end":20,"label":"SKILL"},{"start":189,"end":199,"token_start":28,"token_end":28,"label":"JOB_ROLE"}]}
    ]
    # fmt: on

    return [Example(**example) for example in raw_examples]


def test_rename_labels(test_examples):

    fixed_examples = []
    for e in test_examples:
        example = rename_labels(e, {"SKILL": "skill"})
        fixed_examples.append(example)

    for e in fixed_examples:
        for s in e.spans:
            assert s.label != "SKILL"


def test_fix_annotations(test_examples):
    disparities = get_label_disparities(test_examples, label1="SKILL", label2="JOB_ROLE")
    assert disparities == {"model", "software development engineer"}

    fixed_examples = []
    for e in test_examples:
        corrections = corrections_from_dict({"software development engineer": "JOB_ROLE"})
        example = fix_annotations(e.copy(deep=True), corrections)
        fixed_examples.append(example)

    disparities_fixed = get_label_disparities(fixed_examples, label1="SKILL", label2="JOB_ROLE")
    assert disparities_fixed == {"model"}



from pathlib import Path

import typer
from recon.loaders import read_jsonl
from recon.stats import get_ner_stats


def main(data_file: Path):
    data = read_jsonl(data_file)
    train_stats = get_ner_stats(data)
    print(get_ner_stats(data, serialize=True))


if __name__ == "__main__":
    typer.run(main)



import shutil
from pathlib import Path
from typing import cast

import pytest
import recon
from recon.dataset import Dataset
from recon.operations import registry
from recon.stats import get_ner_stats
from recon.store import ExampleStore
from recon.types import NERStats, OperationStatus, TransformationType


def test_dataset_initialize(example_data):

    dataset = Dataset("train")
    assert dataset.name == "train"
    assert dataset.data == []
    assert dataset.example_store._map == {}
    assert dataset.commit_hash == "94efdd6f628eda9c1ae893467c9652808443ef3e"
    assert dataset.operations == []

    store = ExampleStore()
    dataset2 = Dataset("dev", example_data["dev"], [], store)
    assert dataset2.name == "dev"
    assert dataset2.data == example_data["dev"]
    assert dataset2.example_store == store
    assert dataset2.commit_hash == "dd05e54668c166d075bc4406bfee590e4c89a292"
    assert dataset2.operations == []


def test_dataset_commit_hash(example_data):
    train_dataset = Dataset("train", example_data["train"][:-1])
    dev_dataset = Dataset("train", example_data["dev"])

    assert train_dataset.commit_hash != dev_dataset.commit_hash

    train_commit = train_dataset.commit_hash
    train_dataset.data.append(example_data["train"][-1])

    assert train_dataset.commit_hash != train_commit
    assert hash(train_dataset) == 1186038092183970443


def test_len(example_data):
    train_dataset = Dataset("train", example_data["train"])
    assert len(train_dataset) == len(example_data["train"])


def test_apply(example_data):
    train_dataset = Dataset("train", example_data["train"])

    ner_stats: NERStats = cast(NERStats, train_dataset.apply(get_ner_stats))
    ner_stats_apply: NERStats = cast(NERStats, get_ner_stats(train_dataset.data))

    assert ner_stats.n_examples == ner_stats_apply.n_examples
    assert ner_stats.n_examples_no_entities == ner_stats_apply.n_examples_no_entities
    assert ner_stats.n_annotations == ner_stats_apply.n_annotations
    assert ner_stats.n_annotations_per_type == ner_stats_apply.n_annotations_per_type


def test_apply_(example_data):
    train_dataset = Dataset("train", example_data["train"])
    ner_stats_pre: NERStats = cast(NERStats, train_dataset.apply(get_ner_stats))

    assert len(train_dataset.operations) == 0

    train_dataset.apply_("recon.v1.upcase_labels")

    ner_stats_post: NERStats = cast(NERStats, train_dataset.apply(get_ner_stats))

    pre_keys = sorted(ner_stats_pre.n_annotations_per_type.keys())
    post_keys = sorted(ner_stats_post.n_annotations_per_type.keys())

    assert pre_keys != post_keys

    assert pre_keys == ["JOB_ROLE", "PRODUCT", "SKILL", "product", "skill"]
    assert post_keys == ["JOB_ROLE", "PRODUCT", "SKILL"]

    assert len(train_dataset.operations) == 1

    op = train_dataset.operations[0]

    assert op.name == "recon.v1.upcase_labels"
    assert op.status == OperationStatus.COMPLETED
    assert len(op.transformations) == 3

    for t in op.transformations:
        assert t.type == TransformationType.EXAMPLE_CHANGED


def test_dataset_to_from_disk(example_data, tmp_path):
    train_dataset = Dataset("train", example_data["train"])
    ner_stats_pre: NERStats = cast(NERStats, train_dataset.apply(get_ner_stats))

    assert len(train_dataset.operations) == 0

    with pytest.raises(FileNotFoundError):
        train_dataset.to_disk(tmp_path / "train.jsonl")

    train_dataset.to_disk(tmp_path / "train.jsonl", force=True)
    train_dataset_loaded = Dataset("train").from_disk(tmp_path / "train.jsonl")
    assert len(train_dataset_loaded.operations) == 0
    assert train_dataset_loaded.commit_hash == train_dataset.commit_hash

    train_dataset.apply_("recon.v1.upcase_labels")

    train_dataset.to_disk(tmp_path / "train.jsonl", force=True)
    train_dataset_loaded_2 = Dataset("train").from_disk(tmp_path / "train.jsonl")

    assert len(train_dataset_loaded_2.operations) == 1
    assert train_dataset_loaded_2.commit_hash == train_dataset.commit_hash
    assert train_dataset_loaded_2.commit_hash != train_dataset_loaded.commit_hash

    op = train_dataset_loaded_2.operations[0]

    assert op.name == "recon.v1.upcase_labels"
    assert op.status == OperationStatus.COMPLETED
    assert len(op.transformations) == 3

    for t in op.transformations:
        assert t.type == TransformationType.EXAMPLE_CHANGED



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['recon', 'recon.cli', 'recon.prodigy']

package_data = \
{'': ['*'], 'recon.prodigy': ['templates/*']}

install_requires = \
['pydantic == 1.5', 'spacy >= 2.3.0, <3.0.0', 'scipy', 'typer >= 0.0.8', 'tqdm']

extras_require = \
{'all': ['colorama', 'click-completion'],
 'dev': ['jupyterlab >= 1.2.0, <1.3'],
 'doc': ['mkdocs', 'mkdocs-material', 'markdown-include', 'mkdocstrings'],
 'test': ['pytest == 5.3.0',
          'pytest-cov',
          'coverage',
          'pytest-xdist',
          'pytest-sugar',
          'mypy',
          'black',
          'isort']}

entry_points = \
{'console_scripts': ['recon = recon.cli:app'],
 'prodigy_recipes': ['recon.ner_correct = recon:prodigy_recipes.ner_correct',
                     'recon.ner_merge = recon:prodigy_recipes.ner_merge']}

setup(name='reconner',
      version='0.5.1',
      description='ReconNER, Debug annotated Named Entity Recognition (NER) data for inconsitencies and get insights on improving the quality of your data.',
      author='Kabir Khan',
      author_email='kakh@microsoft.com',
      url='https://github.com/microsoft/reconner',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.6',
     )



import math
from collections import defaultdict
from typing import Any, DefaultDict, Dict, List, Optional, Sequence, Set, Union, cast

import numpy as np
import srsly
from scipy.spatial.distance import jaccard, jensenshannon
from scipy.stats import entropy as scipy_entropy

from .constants import NONE
from .types import (
    EntityCoverage,
    EntityCoverageStats,
    Example,
    NERStats,
    Outliers,
    Span,
)


def get_ner_stats(
    data: List[Example], serialize: bool = False, return_examples: bool = False
) -> Union[NERStats, str, None]:
    """Compute statistics for NER data
    
    Args:
        data (List[Example]): Data as a List of examples
        serialize (bool, optional): Serialize to a JSON string for printing.
        return_examples (bool, optional): Whether to return examples per type
    
    Returns:
        Union[NERStats, str, None]: 
            List of examples or string if serialize and no_print are both True
    """
    annotations_per_type: DefaultDict[str, Any] = defaultdict(int)
    examples: DefaultDict[str, Any] = defaultdict(list)
    n_examples_no_entities = 0
    for e in data:
        if not e.spans:
            n_examples_no_entities += 1
            examples[NONE].append(e)
        else:
            for s in e.spans:
                annotations_per_type[s.label] += 1
                examples[s.label].append(e)

    sorted_anns_by_count = {
        a[0]: a[1] for a in sorted(annotations_per_type.items(), key=lambda x: x[1], reverse=True)
    }

    stats = NERStats(
        n_examples=len(data),
        n_examples_no_entities=n_examples_no_entities,
        n_annotations=sum(annotations_per_type.values()),
        n_annotations_per_type=sorted_anns_by_count,
    )
    if return_examples:
        stats.examples_with_type = examples

    if serialize:
        return srsly.json_dumps(stats.dict(), indent=4)
    else:
        return stats


def get_sorted_type_counts(ner_stats: NERStats) -> List[int]:
    """Get list of counts for each type in n_annotations_per_type property 
    of an NERStats object sorted by type name
    
    Args:
        ner_stats (NERStats): Dataset stats
    
    Returns:
        List[int]: List of counts sorted by type name
    """
    annotations_per_type = ner_stats.n_annotations_per_type
    annotations_per_type[NONE] = ner_stats.n_examples_no_entities

    return [t[1] for t in sorted(annotations_per_type.items(), key=lambda p: p[0])]


def calculate_label_distribution_similarity(x: List[Example], y: List[Example]) -> float:
    """Calculate the similarity of the label distribution for 2 datasets.
    
    e.g. This can help you understand how well your train set models your dev and test sets.
    Empircally you want a similarity over **0.8** when comparing your train set to each of your
    dev and test sets.

        calculate_label_distribution_similarity(corpus.train, corpus.dev)
        # 98.57

        calculate_label_distribution_similarity(corpus.train, corpus.test)
        # 73.29 - This is bad, let's investigate our test set more
    
    Args:
        x (List[Example]): Dataset
        y (List[Example]): Dataset to compare x to
    
    Returns:
        float: Similarity of label distributions
    """

    def pipeline(data: List[Example]) -> Sequence[float]:
        stats = cast(NERStats, get_ner_stats(data))
        sorted_type_counts = get_sorted_type_counts(stats)
        counts_to_probs = get_probs_from_counts(sorted_type_counts)
        return counts_to_probs

    distance = jensenshannon(pipeline(x), pipeline(y))

    return (1 - distance) * 100


def get_entity_coverage(
    data: List[Example],
    sep: str = "||",
    case_sensitive: bool = False,
    return_examples: bool = False,
) -> List[EntityCoverage]:
    """Identify how well you dataset covers an entity type. Get insights
    on the how many times certain text/label span combinations exist across your
    data so that you can focus your annotation efforts better rather than
    annotating examples your Model already understands well.
    
    Args:
        data (List[Example]): List of examples
        sep (str, optional): Separator used in coverage map, only change if || exists in your text
            or label.
        case_sensitive (bool, optional): Consider case of text for each annotation
        return_examples (bool, optional): Return Examples that contain the entity label annotation.
    
    Returns:
        List[EntityCoverage]: Sorted List of EntityCoverage objects containing the text, label, count, and
            an optional list of examples where that text/label annotation exists.
    """
    coverage_map: DefaultDict[str, int] = defaultdict(int)
    examples_map: DefaultDict[str, List[Example]] = defaultdict(list)

    for example in data:
        for span in example.spans:
            text = span.text if case_sensitive else span.text.lower()
            key = f"{text}{sep}{span.label}"
            coverage_map[key] += 1
            examples_map[key].append(example)

    coverage = []
    for key, count in coverage_map.items():
        text, label = key.split(sep)
        record = EntityCoverage(text=text, label=label, count=count)
        if return_examples:
            record.examples = examples_map[key]
        coverage.append(record)

    sorted_coverage = sorted(coverage, key=lambda x: x.count, reverse=True)
    return sorted_coverage


def calculate_entity_coverage_similarity(x: List[Example], y: List[Example]) -> EntityCoverageStats:
    """Calculate how well dataset x covers the entities in dataset y.
    This function should be used to calculate how similar your train set
    annotations cover the annotations in your dev/test set
    
    Args:
        x (List[Example]): Dataset to compare coverage to (usually corpus.train)
        y (List[Example]): Dataset to evaluate coverage for (usually corpus.dev or corpus.test)
    
    Returns:
        EntityCoverageStats: Stats with 
            1. The base entity coverage (does entity in y exist in x)
            2. Count coverage (sum of the EntityCoverage.count property for 
            each EntityCoverage in y to get a more holisic coverage scaled by how 
            often entities occur in each dataset x and y)
    """

    def pipeline(data: List[Example]) -> Dict[int, int]:
        ecs = get_entity_coverage(data)
        return {hash(ec): ec.count for ec in ecs}

    x_map = pipeline(x)
    y_map = pipeline(y)

    n_intersection = 0
    count_intersection = 0
    n_union = 0
    count_union = 0

    for k, count in y_map.items():
        if k in x_map:
            n_intersection += 1
            count_intersection += count
        n_union += 1
        count_union += count

    return EntityCoverageStats(
        entity=(n_intersection / n_union) * 100, count=(count_intersection / count_union) * 100,
    )


def get_probs_from_counts(seq: Sequence[int]) -> Sequence[float]:
    """Convert a sequence of counts to a sequence of probabilties
    by dividing each n by the sum of all n in seq
    
    Args:
        seq (Sequence[int]): Sequence of counts
    
    Returns:
        Sequence[float]: Sequence of probabilities
    """
    return np.asarray(seq) / sum(seq)


def entropy(seq: Union[List[int], List[float]], total: int = None) -> float:
    """Calculate Shannon Entropy for a sequence of Floats or Integers.
    If Floats, check they are probabilities
    If Integers, divide each n in seq by total and calculate entropy
    
    Args:
        seq (Union[List[int], List[float]]): Sequence to calculate entropy for
        total (int, optional): Total to divide by for List of int
    
    Raises:
        ValueError: If seq is not valid
    
    Returns:
        float: Entropy for sequence
    """
    if not seq:
        raise ValueError("Pass a valid non-empty sequence")

    if isinstance(seq[0], float):
        e = scipy_entropy(seq)
    elif isinstance(seq[0], int):
        e = scipy_entropy(get_probs_from_counts(seq))
    else:
        raise ValueError("Parameter seq must be a sequence of probabilites or integers.")
    return e


def calculate_label_balance_entropy(ner_stats: NERStats) -> float:
    """Use Entropy to calculate a metric for label balance based on an NERStats object 
    
    Args:
        ner_stats (NERStats): NERStats for a dataset.
    
    Returns:
        float: Entropy for annotation counts of each label
    """
    total = ner_stats.n_annotations
    classes = [count for label, count in ner_stats.n_annotations_per_type.items()]
    return entropy(classes, total)


def calculate_entity_coverage_entropy(entity_coverage: List[EntityCoverage],) -> float:
    """Use Entropy to calculate a metric for entity coverage.
    
    Args:
        entity_coverage (List[EntityCoverage]): List of EntityCoverage 
            from get_entity_coverage
    
    Returns:
        float: Entropy for entity coverage counts
    """
    counts = [ecs.count for ecs in entity_coverage]
    return entropy(counts, sum(counts))  # type: ignore


def detect_outliers(seq: Sequence[Any], use_log: bool = False) -> Outliers:
    """Detect outliers in a numerical sequence.
    
    Args:
        seq (Sequence[Any]): Sequence of ints or floats
        use_log (bool, optional): Use logarithm of seq.
    
    Returns:
        Tuple[List[int], List[int]]: Tuple of low and high indices
    """
    q1 = np.quantile(seq, 0.25)
    q3 = np.quantile(seq, 0.75)
    iqr = q3 - q1
    fence_low = math.floor(q1 - 1.5 * iqr)
    fence_high = math.floor(q3 + 1.5 * iqr)
    low_indices = [i for i, n in enumerate(seq) if n <= fence_low]
    high_indices = [i for i, n in enumerate(seq) if n > fence_high]
    return Outliers(low=low_indices, high=high_indices)



"""Load a list of Example data from JSON represented records in 
the [Prodigy](https://prodi.gy) format.
"""

import functools
from pathlib import Path
from typing import Any, Dict, List, Union

import srsly

from .operations import registry
from .registry import loading_pipelines
from .types import Example, OperationState, OperationStatus, Transformation


def read_jsonl(path: Path) -> List[Example]:
    """Read annotations in JSONL file format
    
    Args:
        path (Path): Path to data
    
    Returns:
        List[Example]: List of examples
    """
    data = srsly.read_jsonl(path)
    examples = json_to_examples(data)
    return examples


def read_json(path: Path) -> List[Example]:
    """Read annotations in JSON file format
    
    Args:
        path (Path): Path to data
    
    Returns:
        List[Example]: List of examples
    """
    data = srsly.read_jsonl(path)
    examples = json_to_examples(data)
    return examples


def json_to_examples(data: List[Dict[str, Any]]) -> List[Example]:
    """Convert List of Dicts to List of typed Examples
    
    Args:
        data (List[Dict[str, Any]]): Input List of Dicts to convert
    
    Returns:
        List[Example]: List of typed Examples
    """
    return [Example(**example) for example in data]



import copy
from collections import defaultdict
from typing import Any, Dict, List, Set, Tuple, Union

from spacy.language import Language

from .dataset import Dataset
from .operations import op_iter, operation
from .preprocess import SpacyPreProcessor
from .registry import tokenizers
from .types import (
    Example,
    OperationState,
    Span,
    Token,
    Transformation,
    TransformationCallbacks,
    TransformationType,
)

tokenizer = tokenizers.get("default")
nlp = tokenizer()
spacy_pre = SpacyPreProcessor(nlp)


@operation("recon.v1.fix_tokenization_and_spacing", pre=[spacy_pre])
def fix_tokenization_and_spacing(
    example: Example, *, preprocessed_outputs: Dict[str, Any] = {}
) -> Union[Example, None]:
    """Fix tokenization and spacing issues where there are annotation spans that 
    don't fall on a token boundary. This can happen if annotations are done at the
    character level, not the token level. Often, when scraping web text it's easy to
    get two words pushed together where one is an entity so this can fix a lot of issues.
    
    Args:
        example (Example): Input Example
        preprocessed_outputs (Dict[str, Any]): Outputs of preprocessors

    Returns:
        Example: Example with spans fixed to align to token boundaries.
    """

    doc = preprocessed_outputs["recon.v1.spacy"]

    tokens = []
    token_starts = {}
    token_ends = {}

    for t in doc:
        start = t.idx
        end = t.idx + len(t)
        tokens.append(Token(text=t.text, start=start, end=end, id=t.i))
        token_starts[start] = t
        token_ends[end] = t

    spans_to_increment: Dict[int, int] = defaultdict(int)
    for span_i, span in enumerate(example.spans):
        if span.start in token_starts and span.end in token_ends:
            # Aligns to token boundaries, nothing to change here
            continue

        if span.start in token_starts and span.end not in token_ends:
            # Span start aligns to token_start but end doesn't
            # e.g. [customer][PERSONTYPE]s but should be annotated as [customers][PERSONTYPE]
            # tokenization_errors.append((example, span))
            # print("BAD END")
            if span.end + 1 in token_ends:
                # Likely off by 1 annotation
                # e.g. [customer][PERSONTYPE]s but should be annotated as [customers][PERSONTYPE]
                span.end += 1
                span.text = example.text[span.start : span.end]
                # print("SPAN CORRECTED OFF BY 1", example.text, span)
            elif span.end - 1 in token_ends:
                span.end -= 1
                span.text = example.text[span.start : span.end]
            else:
                # Likely bad tokenization
                # e.g. [Quadling][GPE]Country should be split to [Quadling][GPE] Country
                for j in range(span_i + 1, len(example.spans)):
                    spans_to_increment[j] += 1
                fe_text = example.text

                split_start = span.start
                if len(spans_to_increment) > 1 and span_i != list(spans_to_increment.keys())[0]:
                    split_start += spans_to_increment.get(span_i, 0)
                split_end = span.end
                if len(spans_to_increment) > 1 and span_i != list(spans_to_increment.keys())[0]:
                    split_end += spans_to_increment.get(span_i, 0)
                new_text = f"{fe_text[:split_start]}{span.text} {fe_text[split_end:]}"

                example.text = new_text

        elif span.start not in token_starts and span.end in token_ends:
            # Bad tokenization
            # e.g. with[Raymond][PERSON] but text should be split to with [Raymond][PERSON]
            # print("BAD START", span.text)
            # tokenization_errors.append((example, span))
            for j in range(span_i, len(example.spans)):
                spans_to_increment[j] += 1

            fe_text = example.text

            split_start = span.start
            if len(spans_to_increment) > 1 and span_i != list(spans_to_increment.keys())[0]:
                split_start += spans_to_increment.get(span_i, 0)
            split_end = span.end
            if len(spans_to_increment) > 1 and span_i != list(spans_to_increment.keys())[0]:
                split_end += spans_to_increment.get(span_i, 0)

            new_text = f"{fe_text[:split_start]} {span.text}{fe_text[split_end:]}"
            example.text = new_text
        else:
            # Something is super fucked up.
            # print("SPAN CORRECTED OFF BY 1 unfixable", example.text, span)
            before = span.start
            after = span.end
            # tokenization_errors.append((example, span))

            # if (before >= 0 and after < len(span.text) and span[before] not in token_starts and span[before] != ' ' and span[after] not in token_ends and span[after] != ' '):
            #     fe_text = example.text
            #     new_text = f"{fe_text[:span.start]} {span.text}{fe_text[span.end:]}"
            #     spans_to_increment[span_i] += 1
            #     for j in range(span_i + 1, len(example.spans)):
            #         spans_to_increment[j] += 2
            # else:
            # unfixable_examples.add(example.text)
            # break
            return None

    # Increment the start and end characters for each span
    for span_i, count in spans_to_increment.items():
        example.spans[span_i].start += count
        example.spans[span_i].end += count

    return example


@operation("recon.v1.add_tokens", pre=[spacy_pre])
def add_tokens(
    example: Example, *, use_spacy_token_ends: bool = False, preprocessed_outputs: Dict[str, Any]
) -> Union[Example, None]:
    """Add tokens to each Example
    
    Args:
        example (Example): Input Example
        preprocessed_outputs (Dict[str, Any]): Outputs of preprocessors

    Returns:
        Example: Example with tokens
    """
    doc = preprocessed_outputs["recon.v1.spacy"]

    tokens = []
    token_starts = {}
    token_ends = {}

    for t in doc:
        start = t.idx
        end = t.idx + len(t)
        tokens.append(Token(text=t.text, start=start, end=end, id=t.i))
        token_starts[start] = t
        token_ends[end] = t

    example.tokens = tokens

    for span in example.spans:
        if span.start in token_starts and span.end in token_ends:
            span.token_start = token_starts[span.start].i
            if use_spacy_token_ends:
                span.token_end = token_ends[span.end].i + 1
            else:
                span.token_end = token_ends[span.end].i

        if span.token_start is None or span.token_end is None:
            return None

    return example



# isort:skip_file

import copy
from pathlib import Path
from typing import Dict, List

import numpy as np
import pandas as pd
import plotly.express as px
import spacy
import streamlit as st
import typer
from wasabi import Printer
from recon import Corpus
from recon.constants import NONE
from recon.corrections import rename_labels
from recon.insights import (
    get_ents_by_label,
    get_hardest_examples,
    get_label_disparities,
    top_label_disparities,
    top_prediction_errors,
)
from recon.recognizer import SpacyEntityRecognizer
from recon.stats import get_entity_coverage, get_ner_stats
from recon.types import Example, HardestExample, PredictionError


@st.cache(allow_output_mutation=True)
def load_model(name):
    return spacy.load(name)

@st.cache
def load_dataset(data_dir: Path):
    return Corpus.from_disk(data_dir)


st.sidebar.title("Interactive Recon Visualizer")
st.sidebar.markdown(
    """
Process your NER data with [Recon](https://microsoft.github.io/reconner)
"""
)

spacy_model = st.sidebar.text_input("Model name or path")
dataset_dir = st.sidebar.text_input("Data Directory")

nlp = load_model(spacy_model)
ds = load_dataset(dataset_dir)

st.header("Stats")
for name, stats in ds.apply(get_ner_stats).items():
    if 'examples_with_type' in stats:
        del stats['examples_with_type']
    st.subheader(f"{name} data - {stats['n_examples']} Total Examples")
    ents_per_type = stats['ents_per_type']

    chart_data = pd.DataFrame(
        np.asarray(list(zip(ents_per_type.keys(), ents_per_type.values()))),
        columns=["Label", "N Annotations"]
    ).sort_values("Label")
    fig = px.bar(chart_data, x="N Annotations", y="Label", orientation="h")
    st.plotly_chart(fig)
    st.json(stats)



# def main(model: str = typer.Option(...), data_dir: Path = typer.Option(...)):
#     nlp = load_model(model)
#     ds = load_dataset(data_dir)

#     st.text(get_ner_stats(ds.test, serialize=True, no_print=True))


# if __name__ == "__main__":
#     typer.run(main)



"""Make corrections to your data."""

import copy
from collections import defaultdict
from typing import Any, DefaultDict, Dict, List

import spacy
from spacy.tokens import Doc as SpacyDoc, Span as SpacySpan
from wasabi import msg

from .operations import operation
from .preprocess import SpacyPreProcessor
from .types import Correction, Example, Span, Token, TransformationCallbacks


@operation("recon.v1.rename_labels")
def rename_labels(example: Example, label_map: Dict[str, str]) -> Example:
    """Rename labels in a copy of List[Example] data
    
    Args:
        example (Example): Input Example
        label_map (Dict[str, str]): One-to-one mapping of label names
    
    Returns:
        Example: Copy of Example with renamed labels
    """
    for span in example.spans:
        span.label = label_map.get(span.label, span.label)
    return example


@operation("recon.v1.fix_annotations")
def fix_annotations(
    example: Example,
    corrections: List[Correction],
    case_sensitive: bool = False,
    dryrun: bool = False,
) -> Example:
    """Fix annotations in a copy of List[Example] data.
    
    This function will NOT add annotations to your data.
    It will only remove erroneous annotations and fix the
    labels for specific spans.
    
    Args:
        example (Example): Input Example
        corrections (Dict[str, str]): Dictionary of corrections mapping entity text to a new label.
            If the value is set to None, the annotation will be removed
        case_sensitive (bool, optional): Consider case of text for each correction
        dryrun (bool, optional): Treat corrections as a dryrun and just print all changes to be made
    
    Returns:
        Example: Example with fixed annotations
    """

    if not case_sensitive:
        for c in corrections:
            c.annotation = c.annotation.lower()

    corrections_map: Dict[str, Correction] = {c.annotation: c for c in corrections}
    prints: List[str] = []

    ents_to_remove: List[int] = []
    for i, s in enumerate(example.spans):
        t = s.text if case_sensitive else s.text.lower()

        if t in corrections_map:
            c = corrections_map[t]
            if c.to_label is None and s.label == c.from_label:
                if dryrun:
                    prints.append(f"Deleting span: {s.text}")
                else:
                    ents_to_remove.append(i)
            elif s.label in c.from_labels or "ANY" in c.from_labels:
                if dryrun:
                    prints.append(
                        f"Correction span: {s.text} from labels: {c.from_labels} to label: {c.to_label}"
                    )
                else:
                    s.label = c.to_label

    i = len(ents_to_remove) - 1
    while i >= 0:
        idx = ents_to_remove[i]
        del example.spans[idx]
        i -= 1

    if dryrun:
        msg.divider("Example Text")
        msg.text(example.text)
        for line in prints:
            msg.text(line)

    return example


def corrections_from_dict(corrections_dict: Dict[str, Any]) -> List[Correction]:
    """Create a list of Correction objects from a simpler config for
    corrections using a Dict representation mapping keys to either the label to 
    convert to or a tuple of (from_label, to_label) pairings or (List[from_labels], to_label)
    pairings if you want to convert as subset of labels at a time

    Args:
        corrections_dict (Dict[str, Any]): Corrections formatted dict
            e.g. {
                "united states": "GPE",
                "London": (["LOC"], "GPE")
            }

    Raises:
        ValueError: If the format of the dict

    Returns:
        [type]: [description]
    """
    corrections: List[Correction] = []
    for key, val in corrections_dict.items():
        if isinstance(val, str) or val == None:
            from_labels = ["ANY"]
            to_label = val
        elif isinstance(val, tuple):
            if isinstance(val[0], str):
                from_labels = [val[0]]
            else:
                from_labels = val[0]
            to_label = val[1]
        else:
            raise ValueError(
                "Cannot parse corrections dict. Value must be either a str of the label "
                + "to change the annotation to (TO_LABEL) or a tuple of (FROM_LABEL, TO_LABEL)"
            )
        corrections.append(Correction(annotation=key, from_labels=from_labels, to_label=to_label))
    return corrections


@operation("recon.v1.strip_annotations")
def strip_annotations(
    example: Example, strip_chars: List[str] = [".", "!", "?", "-", ":", " "]
) -> Example:
    """Strip punctuation and spaces from start and end of annotations.
    These characters are almost always a mistake and will confuse a model
    
    Args:
        example (Example): Input Example
        strip_chars (List[str], optional): Characters to strip.
    
    Returns:
        Example: Example with stripped spans
    """

    for s in example.spans:
        for ch in strip_chars:
            if s.text.startswith(ch):
                ch = s.text[0]

                while ch in strip_chars:
                    s.text = s.text[1:]
                    s.start += 1
                    ch = s.text[0]
            elif s.text.endswith(ch):
                ch = s.text[-1]
                while ch in strip_chars:
                    s.text = s.text[:-1]
                    ch = s.text[-1]
                    s.end -= 1
    return example


nlp = spacy.blank("en")
nlp.add_pipe(nlp.create_pipe("sentencizer"))
spacy_pre = SpacyPreProcessor(nlp)


@operation("recon.v1.split_sentences", pre=[spacy_pre])
def split_sentences(example: Example, preprocessed_outputs: Dict[str, Any] = {}) -> List[Example]:
    """Split a single example into multiple examples by splitting the text into 
    multiple sentences and resetting entity and token offsets based on offsets 
    relative to sentence boundaries

    Args:
        example (Example): Input Example
        preprocessed_outputs (Dict[str, Any], optional): Outputs of preprocessors.

    Returns:
        List[Example]: List of split examples. 
            Could be list of 1 if the example is just one sentence.
    """
    doc = preprocessed_outputs["recon.v1.spacy"]

    new_examples = []
    ents = []
    for ent in example.spans:
        span = doc.char_span(ent.start, ent.end, label=ent.label)
        if not span:
            token = None
            text = doc.text[ent.start : ent.end]
            for t in doc:
                if t.text == text:
                    token = t
            if token:
                span = SpacySpan(doc, token.i, token.i + 1, label=ent.label)
        ents.append(span)

    doc.ents = ents

    for sent in doc.sents:
        sent_doc = sent.as_doc()
        new_example = Example(
            text=sent_doc.text,
            spans=[
                Span(
                    text=e.text,
                    start=e.start_char,
                    end=e.end_char,
                    token_start=e.start,
                    token_end=e.end,
                    label=e.label_,
                )
                for e in sent_doc.ents
            ],
            tokens=[
                Token(text=t.text, start=t.idx, end=t.idx + len(t.text), id=i)
                for i, t in enumerate(sent_doc)
            ],
        )
        new_examples.append(new_example)
    return new_examples
